from typing import List 
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.config import Config


class _Config:
    def __init__(self, api_obj):
        self.AddDnValue = self._AddDnValue(api_obj=api_obj)
        self.AddPolicyValue = self._AddPolicyValue(api_obj=api_obj)
        self.AddValue = self._AddValue(api_obj=api_obj)
        self.ClearAttribute = self._ClearAttribute(api_obj=api_obj)
        self.ClearPolicyAttribute = self._ClearPolicyAttribute(api_obj=api_obj)
        self.ContainableClasses = self._ContainableClasses(api_obj=api_obj)
        self.Create = self._Create(api_obj=api_obj)
        self.DefaultDN = self._DefaultDN(api_obj=api_obj)
        self.Delete = self._Delete(api_obj=api_obj)
        self.DnToGuid = self._DnToGuid(api_obj=api_obj)
        self.Enumerate = self._Enumerate(api_obj=api_obj)
        self.EnumerateAll = self._EnumerateAll(api_obj=api_obj)
        self.EnumerateObjectsDerivedFrom = self._EnumerateObjectsDerivedFrom(api_obj=api_obj)
        self.EnumeratePolicies = self._EnumeratePolicies(api_obj=api_obj)
        self.Find = self._Find(api_obj=api_obj)
        self.FindObjectsOfClass = self._FindObjectsOfClass(api_obj=api_obj)
        self.FindPolicy = self._FindPolicy(api_obj=api_obj)
        self.GetHighestRevision = self._GetHighestRevision(api_obj=api_obj)
        self.GetRevision = self._GetRevision(api_obj=api_obj)
        self.GuidToDn = self._GuidToDn(api_obj=api_obj)
        self.IdInfo = self._IdInfo(api_obj=api_obj)
        self.IsValid = self._IsValid(api_obj=api_obj)
        self.MutateObject = self._MutateObject(api_obj=api_obj)
        self.Read = self._Read(api_obj=api_obj)
        self.ReadAll = self._ReadAll(api_obj=api_obj)
        self.ReadDn = self._ReadDn(api_obj=api_obj)
        self.ReadDnReferences = self._ReadDnReferences(api_obj=api_obj)
        self.ReadEffectivePolicy = self._ReadEffectivePolicy(api_obj=api_obj)
        self.ReadPolicy = self._ReadPolicy(api_obj=api_obj)
        self.RemoveDnValue = self._RemoveDnValue(api_obj=api_obj)
        self.RemovePolicyValue = self._RemovePolicyValue(api_obj=api_obj)
        self.RenameObject = self._RenameObject(api_obj=api_obj)
        self.Write = self._Write(api_obj=api_obj)
        self.WriteDn = self._WriteDn(api_obj=api_obj)
        self.WritePolicy = self._WritePolicy(api_obj=api_obj)

    class _AddDnValue(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/AddDnValue')

        def post(self, object_dn: str, attribute_name: str, value: str):
            body = {
                'ObjectDN': object_dn,
                'AttributeName': attribute_name,
                'Value': value
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
                    
            return _Response(response=self._post(data=body))

    class _AddPolicyValue(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/AddPolicyValue')

        def post(self, object_dn: str, attribute_name: str, class_name: str, value: str, locked: bool):
            body = {
                'ObjectDN': object_dn,
                'AttributeName': attribute_name,
                'Class': class_name,
                'Value': value,
                'Locked': locked
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _AddValue(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/AddValue')

        def post(self, object_dn: str, attribute_name: str, value: str):
            body = {
                'ObjectDN': object_dn,
                'AttributeName': attribute_name,
                'Value': value,
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _ClearAttribute(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/ClearAttribute')

        def post(self, object_dn: str, attribute_name: str):
            body = {
                'ObjectDN': object_dn,
                'AttributeName': attribute_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _ClearPolicyAttribute(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/ClearPolicyAttribute')

        def post(self, object_dn: str, class_name: str, attribute_name: str):
            body = {
                'ObjectDN': object_dn,
                'Class': class_name,
                'AttributeName': attribute_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _ContainableClasses(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/ContainableClasses')

        def post(self, object_dn: str):
            body = {
                'ObjectDN': object_dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def class_names(self) -> List[str]:
                    return self._from_json(key='ClassNames')

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _CountObjects(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/CountObjects')

        def post(self, object_dn: str, type_name: str, recursive: bool = False, pattern: str = None):
            body = {
                'ObjectDN': object_dn,
                'Type': type_name,
                'Pattern': pattern,
                'Recursive': recursive
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def count(self) -> int:
                    return self._from_json(key='Count')

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _Create(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/Create')

        def post(self, object_dn: str, class_name: str, name_attribute_list: list):
            body = {
                "ObjectDN": object_dn,
                "Class": class_name,
                "NameAttributeList": name_attribute_list
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def object(self):
                    return Config.Object(self._from_json(key='Object'))

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _DefaultDN(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/DefaultDN')

        def post(self, default_dn: str):
            body = {
                'DefaultDN': default_dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def default_dn(self) -> str:
                    return self._from_json(key='DefaultDN')

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _Delete(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/Delete')

        def post(self, object_dn: str, recursive: bool = False):
            body = {
                "ObjectDN": object_dn,
                "Recursive": recursive
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _DnToGuid(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/DnToGuid')

        def post(self, object_dn: str):
            body = {
                "ObjectDN": object_dn,
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def class_name(self) -> str:
                    return self._from_json(key='ClassName')

                @property
                @api_response_property()
                def guid(self) -> str:
                    return self._from_json(key='GUID')

                @property
                @api_response_property()
                def revision(self) -> str:
                    return self._from_json(key='Revision')

                @property
                @api_response_property()
                def hierarchical_guid(self) -> str:
                    return self._from_json(key='HierarchicalGUID')

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _Enumerate(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/Enumerate')

        def post(self, object_dn: str = None, recursive: bool = False, pattern: str = None):
            body = {
                "ObjectDN": object_dn,
                "Recursive": recursive,
                "Pattern": pattern
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def objects(self):
                    return [Config.Object(obj) for obj in self._from_json(key='Objects')]

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _EnumerateAll(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/EnumerateAll')

        def post(self, pattern: str):
            body = {
                "Pattern": pattern
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def objects(self):
                    return [Config.Object(obj) for obj in self._from_json(key='Objects')]

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _EnumerateObjectsDerivedFrom(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/EnumerateObjectsDerivedFrom')

        def post(self, derived_from: str, pattern: str = None, object_dn: str = None):
            body = {
                "ObjectDN": object_dn,
                "DerivedFrom": derived_from,
                "Pattern": pattern
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def objects(self):
                    return [Config.Object(obj) for obj in self._from_json(key='Objects')]

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _EnumeratePolicies(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/EnumeratePolicies')

        def post(self, object_dn: str):
            body = {
                "ObjectDN": object_dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def policies(self):
                    return [Config.Policy(obj) for obj in self._from_json(key='Policies')]

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _Find(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/Find')

        def post(self, pattern: str, attribute_names: str = None):
            body = {
                "Pattern": pattern,
                "AttributeNames": attribute_names
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def objects(self):
                    return [Config.Object(obj) for obj in self._from_json(key='Objects')]

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _FindContainers(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/FindContainers')

        def post(self, object_dn: str, recursive: bool = False):
            body = {
                "ObjectDN": object_dn,
                "Recursive": recursive
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def objects(self):
                    return [Config.Object(obj) for obj in self._from_json(key='Objects')]

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _FindObjectsOfClass(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/FindObjectsOfClass')

        def post(self, classes: str = None, class_name: str = None, object_dn: str = None, pattern: str = None, recursive: bool = False):
            if not (classes or class_name):
                raise AssertionError('One of "classes" or "class_name" parameters must be provided.')
            body = {
                "Classes": classes,
                "Class": class_name,
                'ObjectDN': object_dn,
                'Pattern': pattern,
                'Recursive': recursive
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def objects(self):
                    return [Config.Object(obj) for obj in self._from_json(key='Objects')]

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _FindPolicy(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/FindPolicy')

        def post(self, object_dn: str, class_name: str, attribute_name: str):
            body = {
                "ObjectDN": object_dn,
                "Class": class_name,
                "AttributeName": attribute_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json(key='Locked')

                @property
                @api_response_property()
                def policy_dn(self) -> str:
                    return self._from_json(key='PolicyDN')

                @property
                @api_response_property()
                def values(self) -> List[str]:
                    return self._from_json(key='Values')

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _GetHighestRevision(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/GetHighestRevision')

        def post(self, object_dn: str, classes: str = None):
            body = {
                "ObjectDN": object_dn,
                'Classes': classes
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def revision(self) -> str:
                    return self._from_json(key='Revision')

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _GetRevision(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/GetRevision')

        def post(self, object_dn: str):
            body = {
                "ObjectDN": object_dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def revision(self) -> str:
                    return self._from_json(key='Revision')

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _GuidToDn(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/GuidToDn')

        def post(self, object_guid: str):
            body = {
                "ObjectGUID": object_guid
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def object_dn(self) -> str:
                    return self._from_json(key='ObjectDN')

                @property
                @api_response_property()
                def class_name(self) -> str:
                    return self._from_json(key='ClassName')

                @property
                @api_response_property()
                def revision(self) -> str:
                    return self._from_json(key='Revision')

                @property
                @api_response_property()
                def hierarchical_guid(self) -> str:
                    return self._from_json(key='HierarchicalGUID')

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _IdInfo(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/IdInfo')

        def post(self, object_id: str):
            body = {
                "ObjectID": object_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def guid(self) -> str:
                    return self._from_json(key='GUID')

                @property
                @api_response_property()
                def class_name(self) -> str:
                    return self._from_json(key='ClassName')

                @property
                @api_response_property()
                def revision(self) -> str:
                    return self._from_json(key='Revision')

                @property
                @api_response_property()
                def hierarchical_guid(self) -> str:
                    return self._from_json(key='HierarchicalGUID')

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _IsValid(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/IsValid')

        def post(self, object_dn: str = None, object_guid: str = None):
            if not (object_dn or object_guid):
                raise AssertionError('One of "classes" or "class_name" parameters must be provided.')
            body = {
                "ObjectGUID": object_guid,
                "ObjectDN": object_dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def object(self):
                    return Config.Object(self._from_json(key='Object'))

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _MutateObject(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/MutateObject')

        def post(self, object_dn: str, class_name: str):
            body = {
                "ObjectDN": object_dn,
                "Class": class_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _Read(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/Read')

        def post(self, object_dn: str, attribute_name: str):
            body = {
                "ObjectDN": object_dn,
                "AttributeName": attribute_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def object_dn(self) -> str:
                    return self._from_json(key='ObjectDN')

                @property
                @api_response_property()
                def attribute_name(self) -> str:
                    return self._from_json(key='AttributeName')

                @property
                @api_response_property()
                def values(self) -> List[str]:
                    return self._from_json(key='Values')

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _ReadAll(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/ReadAll')

        def post(self, object_dn: str):
            body = {
                "ObjectDN": object_dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def name_values(self):
                    return [Config.NameValues(nv) for nv in self._from_json(key='NameValues')]

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _ReadDn(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/ReadDn')

        def post(self, object_dn: str, attribute_name: str):
            body = {
                "ObjectDN": object_dn,
                "AttributeName": attribute_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def values(self) -> List[str]:
                    return self._from_json(key='Values')

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _ReadDnReferences(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/ReadDnReferences')

        def post(self, object_dn: str, reference_attribute_name: str, attribute_name: str):
            body = {
                "ObjectDN": object_dn,
                "ReferenceAttributeName": reference_attribute_name,
                "AttributeName": attribute_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def values(self) -> List[str]:
                    return self._from_json(key='Values', return_on_error=list)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _ReadEffectivePolicy(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/ReadEffectivePolicy')

        def post(self, object_dn: str, attribute_name: str):
            body = {
                "ObjectDN": object_dn,
                "AttributeName": attribute_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def values(self) -> List[str]:
                    return self._from_json(key='Values')

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json(key='Locked')

                @property
                @api_response_property()
                def overridden(self) -> bool:
                    return self._from_json(key='Overridden')

                @property
                @api_response_property()
                def policy_dn(self) -> str:
                    return self._from_json(key='PolicyDN')

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _ReadPolicy(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/ReadPolicy')

        def post(self, object_dn: str, attribute_name: str, class_name: str):
            body = {
                "ObjectDN": object_dn,
                "AttributeName": attribute_name,
                "Class": class_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json(key='Locked')

                @property
                @api_response_property()
                def values(self) -> List[str]:
                    return self._from_json(key='Values', return_on_error=list)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _RemoveDnValue(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/RemoveDnValue')

        def post(self, object_dn: str, attribute_name: str, value: str):
            body = {
                "ObjectDN": object_dn,
                "AttributeName": attribute_name,
                "Value": value
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))                    
            
            return _Response(response=self._post(data=body))

    class _RemovePolicyValue(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/RemovePolicyValue')

        def post(self, object_dn: str, attribute_name: str, class_name: str, value: str):
            body = {
                "ObjectDN": object_dn,
                "AttributeName": attribute_name,
                "Class": class_name,
                "Value": value
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _RenameObject(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/RenameObject')

        def post(self, object_dn: str, new_object_dn: str):
            body = {
                "ObjectDN": object_dn,
                "NewObjectDN": new_object_dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _Write(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/Write')

        def post(self, object_dn: str, attribute_data: dict):
            body = {
                "ObjectDN": object_dn,
                "AttributeData": attribute_data
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _WriteDn(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/WriteDn')

        def post(self, object_dn: str, attribute_name: str, values: List[str]):
            body = {
                "ObjectDN": object_dn,
                "AttributeName": attribute_name,
                "Values": values
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))

    class _WritePolicy(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Config/WritePolicy')

        def post(self, object_dn: str, class_name: str, attribute_name: str, locked: bool = False, values: str = None):
            body = {
                "ObjectDN": object_dn,
                "Class": class_name,
                "AttributeName": attribute_name,
                "Locked": locked,
                "Values": values
            }
            
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Config.Result(self._from_json(key='Result'))
            
            return _Response(response=self._post(data=body))



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.driver_base import DriverBaseAttributes


class EncryptionDriverAttributes(DriverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Encryption Driver"
	encryption_data = Attribute('Encryption Data')



from typing import List
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.pki import PKI


class _PKI:
    def __init__(self, api_obj):
        self.HashiCorp = self._HashiCorp(api_obj=api_obj)

    class _HashiCorp:
        def __init__(self, api_obj):
            self.CA = self._CA(api_obj=api_obj)
            self.Role = self._Role(api_obj=api_obj)

        class _CA(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/HashiCorp/CA')
                self._api_obj = api_obj

            def Guid(self, guid: str):
                return self._Guid(guid=guid, api_obj=self._api_obj)

            def get(self):
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def pkis(self):
                        return [PKI.PKI(pki) for pki in self._from_json(key='pkis')]

                return _Response(response=self._get())

            def post(self, certificate: dict, folder_dn: str, pki_path: str, roles: List[str],
                     create_certificate_authority: bool = True, create_pki_role: bool = False, crl_address: str = None,
                     installation: dict = None, key_algorithm: str = None, key_bit_size: str = None, ocsp_address: str = None):
                body = {
                    'Certificate': certificate,
                    'CreateCertificateAuthority': create_certificate_authority,
                    'CreatePKIRole': create_pki_role,
                    'CRLAddress': crl_address,
                    'FolderDn': folder_dn,
                    'Installation': installation,
                    'KeyAlgorithm': key_algorithm,
                    'KeyBitSize': key_bit_size,
                    'OCSPAddress': ocsp_address,
                    'PkiPath': pki_path,
                    'Roles': roles
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def certificate_dn(self) -> str:
                        return self._from_json(key='CertificateDN')

                    @property
                    @api_response_property()
                    def certificate_guid(self) -> str:
                        return self._from_json(key='CertificateGuid')

                    @property
                    @api_response_property()
                    def error(self) -> str:
                        return self._from_json(key='Error')

                    @property
                    @api_response_property()
                    def guid(self) -> str:
                        return self._from_json(key='Guid')

                return _Response(response=self._post(data=body))

            class _Guid(API):
                def __init__(self, guid: str, api_obj):
                    super().__init__(api_obj=api_obj, url=f'/HashiCorp/CA/{guid}')
                    self._guid = guid
                    self.Renew = self._Renew(guid=guid, api_obj=api_obj)

                def delete(self):
                    class _Response(APIResponse):
                        def __init__(self, response):
                            super().__init__(response=response)

                    return _Response(response=self._delete())

                def get(self):
                    class _Response(APIResponse):
                        def __init__(self, response):
                            super().__init__(response=response)

                        @property
                        @api_response_property()
                        def certificate(self):
                            return PKI.Certificate(self._from_json(key='Certificate'))

                        @property
                        @api_response_property()
                        def create_certificate_authority(self) -> bool:
                            return self._from_json(key='CreateCertificateAuthority')

                        @property
                        @api_response_property()
                        def create_pki_role(self) -> bool:
                            return self._from_json(key='CreatePKIRole')

                        @property
                        @api_response_property()
                        def folder_dn(self):
                            return self._from_json(key='FolderDn')

                        @property
                        @api_response_property()
                        def installation(self):
                            return PKI.Installation(self._from_json(key='Installation'))

                        @property
                        @api_response_property()
                        def pki_path(self):
                            return self._from_json(key='PkiPath')

                        @property
                        @api_response_property()
                        def roles(self) -> list:
                            return self._from_json(key='Roles')

                    return _Response(response=self._get())

                def post(self):
                    class _Response(APIResponse):
                        def __init__(self, response):
                            super().__init__(response=response)

                        @property
                        @api_response_property()
                        def certificate_dn(self) -> str:
                            return self._from_json(key='CertificateDN')

                        @property
                        @api_response_property()
                        def certificate_guid(self) -> str:
                            return self._from_json(key='CertificateGuid')

                        @property
                        @api_response_property()
                        def error(self) -> str:
                            return self._from_json(key='Error')

                        @property
                        @api_response_property()
                        def guid(self) -> str:
                            return self._from_json(key='Guid')

                    return _Response(response=self._post(data={}))

                def put(self, folder_dn: str, pki_path: str, roles: List[str], certificate: dict = None,
                        create_certificate_authority: bool = True, create_pki_role: bool = False, crl_address: str = None,
                        installation: dict = None, key_algorithm: str = None, key_bit_size: str = None, ocsp_address: str = None):
                    body = {
                        'Certificate': certificate,
                        'CreateCertificateAuthority': create_certificate_authority,
                        'CreatePKIRole': create_pki_role,
                        'CRLAddress': crl_address,
                        'FolderDn': folder_dn,
                        'Installation': installation,
                        'KeyAlgorithm': key_algorithm,
                        'KeyBitSize': key_bit_size,
                        'OCSPAddress': ocsp_address,
                        'PkiPath': pki_path,
                        'Roles': roles
                    }

                    class _Response(APIResponse):
                        def __init__(self, response):
                            super().__init__(response=response)

                        @property
                        @api_response_property()
                        def certificate_dn(self) -> str:
                            return self._from_json(key='CertificateDN')

                        @property
                        @api_response_property()
                        def certificate_guid(self) -> str:
                            return self._from_json(key='CertificateGuid')

                        @property
                        @api_response_property()
                        def error(self) -> str:
                            return self._from_json(key='Error')

                        @property
                        @api_response_property()
                        def guid(self) -> str:
                            return self._from_json(key='Guid')

                    return _Response(response=self._put(data=body))

                class _Renew(API):
                    def __init__(self, guid:str, api_obj):
                        super().__init__(api_obj=api_obj, url=f'HashiCorp/{guid}/Renew')

                    def post(self):
                        class _Response(APIResponse):
                            def __init__(self, response):
                                super().__init__(response=response)

                            @property
                            @api_response_property()
                            def certificate_dn(self) -> str:
                                return self._from_json(key='CertificateDN')

                            @property
                            @api_response_property()
                            def certificate_guid(self) -> str:
                                return self._from_json(key='CertificateGuid')

                            @property
                            @api_response_property()
                            def guid(self) -> str:
                                return self._from_json(key='Guid')

                        return _Response(response=self._post(data={}))

        class _Role(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='HashiCorp/Role')
                self._api_obj = api_obj

            def post(self, folder_dn: str, role_name: str, city: str = None, country: str = None,
                     enhanced_key_usage: List[str] = None, key_algorithm: str = None, key_bit_size: str = None,
                     organization: str = None, organizational_units: List[str] = None, state: str = None,
                     whitelisted_domains: List[str] = None):
                body = {
                    'City': city,
                    'Country': country,
                    'EnhancedKeyUsage': enhanced_key_usage,
                    'FolderDn': folder_dn,
                    'KeyAlgorithm': key_algorithm,
                    'KeyBitSize': key_bit_size,
                    'Organization': organization,
                    'OrganizationalUnits': organizational_units,
                    'RoleName': role_name,
                    'State': state,
                    'WhitelistedDomains': whitelisted_domains
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def error(self) -> str:
                        return self._from_json(key='Error')

                    @property
                    @api_response_property()
                    def guid(self) -> str:
                        return self._from_json(key='Guid')

                return _Response(response=self._post(data=body))

            def Guid(self, guid: str):
                return self._Guid(guid=guid, api_obj=self._api_obj)

            class _Guid(API):
                def __init__(self, guid: str, api_obj):
                    super().__init__(api_obj=api_obj, url=f'HashiCorp/Role/{guid}')

                def delete(self):
                    class _Response(APIResponse):
                        def __init__(self, response):
                            super().__init__(response=response)

                    return _Response(response=self._delete())

                def get(self):
                    class _Response(APIResponse):
                        def __init__(self, response):
                            super().__init__(response=response)

                        @property
                        @api_response_property()
                        def city(self) -> str:
                            return self._from_json(key='City')

                        @property
                        @api_response_property()
                        def country(self) -> str:
                            return self._from_json(key='Country')

                        @property
                        @api_response_property()
                        def enhanced_key_usage(self) -> str:
                            return self._from_json(key='EnhancedKeyUsage')

                        @property
                        @api_response_property()
                        def error(self) -> str:
                            return self._from_json(key='Error')

                        @property
                        @api_response_property()
                        def folder_dn(self) -> str:
                            return self._from_json(key='FolderDn')

                        @property
                        @api_response_property()
                        def guid(self) -> str:
                            return self._from_json(key='Guid')

                        @property
                        @api_response_property()
                        def key_algorithm(self) -> str:
                            return self._from_json(key='KeyAlgorithm')

                        @property
                        @api_response_property()
                        def key_bit_size(self) -> str:
                            return self._from_json(key='KeyBitSize')

                        @property
                        @api_response_property()
                        def organization(self) -> str:
                            return self._from_json(key='Organization')

                        @property
                        @api_response_property()
                        def organizational_units(self) -> List[str]:
                            return self._from_json(key='OrganizationalUnits')

                        @property
                        @api_response_property()
                        def role_name(self) -> str:
                            return self._from_json(key='RoleName')

                        @property
                        @api_response_property()
                        def state(self) -> str:
                            return self._from_json(key='State')

                        @property
                        @api_response_property()
                        def whitelisted_domains(self) -> List[str]:
                            return self._from_json(key='WhitelistedDomains')

                    return _Response(response=self._get())

                def put(self, city: str = None, country: str = None,
                        enhanced_key_usage: List[str] = None, key_algorithm: str = None, key_bit_size: str = None,
                        organization: str = None, organizational_units: List[str] = None, state: str = None,
                        whitelisted_domains: List[str] = None):
                    body = {
                        'City': city,
                        'Country': country,
                        'EnhancedKeyUsage': enhanced_key_usage,
                        'KeyAlgorithm': key_algorithm,
                        'KeyBitSize': key_bit_size,
                        'Organization': organization,
                        'OrganizationalUnits': organizational_units,
                        'State': state,
                        'WhitelistedDomains': whitelisted_domains
                    }

                    class _Response(APIResponse):
                        def __init__(self, response):
                            super().__init__(response=response)

                        @property
                        @api_response_property()
                        def error(self) -> str:
                            return self._from_json(key='Error')

                        @property
                        @api_response_property()
                        def guid(self) -> str:
                            return self._from_json(key='Guid')

                    return _Response(response=self._put(data=body))



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_authority_base import CertificateAuthorityBaseAttributes


class UniCERTCAAttributes(CertificateAuthorityBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "UniCERT CA"
	ca_dn = Attribute('CA DN')
	ra_dn = Attribute('RA DN')
	secure = Attribute('Secure')
	web_instance = Attribute('Web Instance')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_trust_store_base import CertificateTrustStoreBaseAttributes


class GSKTrustStoreAttributes(CertificateTrustStoreBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "GSK Trust Store"
	fips_key = Attribute('Fips Key', min_version='15.2')
	key_store = Attribute('Key Store', min_version='15.2')
	key_store_credential = Attribute('Key Store Credential', min_version='15.2')
	stash_password = Attribute('Stash Password', min_version='15.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.code_signing_environment_base import CodeSigningEnvironmentBaseAttributes


class CodeSigningAppleEnvironmentAttributes(CodeSigningEnvironmentBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing Apple Environment"
	cn_pattern = Attribute('CN Pattern', min_version='21.2')
	certificate_dn = Attribute('Certificate DN', min_version='21.2')
	key_storage_location = Attribute('Key Storage Location', min_version='21.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.log_channel import LogChannelAttributes


class LogSyslogAttributes(LogChannelAttributes, metaclass=IterableMeta):
	__config_class__ = "Log Syslog"
	credential = Attribute('Credential', min_version='19.3')
	enable_tls = Attribute('Enable TLS', min_version='19.3')
	facility = Attribute('Facility')
	message_format = Attribute('Message Format', min_version='19.3')
	message_prefix = Attribute('Message Prefix', min_version='19.3')
	port = Attribute('Port', min_version='19.3')
	protocol = Attribute('Protocol', min_version='19.3')
	target = Attribute('Target')
	timeout = Attribute('Timeout', min_version='19.3')



from pytpp.tools.helpers.date_converter import from_date_string
from pytpp.properties.response_objects.dataclasses import log


class Log:
    @staticmethod
    def LogEvent(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return log.LogEvent(
            client_timestamp=from_date_string(response_object.get('ClientTimestamp')),
            component=response_object.get('Component'),
            component_id=response_object.get('ComponentId'),
            component_subsystem=response_object.get('ComponentSubsystem'),
            data=response_object.get('Data'),
            grouping=response_object.get('Grouping'),
            id=response_object.get('Id'),
            name=response_object.get('Name'),
            server_timestamp=from_date_string(response_object.get('ServerTimestamp')),
            severity=response_object.get('Severity'),
            source_ip=response_object.get('SourceIP'),
            text1=response_object.get('Text1'),
            text2=response_object.get('Text2'),
            value1=response_object.get('Value1'),
            value2=response_object.get('Value2'),
        )

    @staticmethod
    def LogEventApplicationDefinition(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return log.LogEventApplicationDefinition(
            application_name=response_object.get('ApplicationName'),
            id=response_object.get('ID'),
        )

    @staticmethod
    def LogEventDefinition(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return log.LogEventDefinition(
            data_format=response_object.get('DataFormat'),
            data_title=response_object.get('DataTitle'),
            description=response_object.get('Description'),
            grouping_title=response_object.get('GroupingTitle'),
            grouping_type=response_object.get('GroupingType'),
            id=response_object.get('ID'),
            text1_title=response_object.get('Text1Title'),
            text2_title=response_object.get('Text2Title'),
            value1_title=response_object.get('Value1Title'),
            value1_type=response_object.get('Value1Type'),
            value2_title=response_object.get('Value2Title'),
            value2_type=response_object.get('Value2Type'),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.schedule_base import ScheduleBaseAttributes
from pytpp.attributes.service_module import ServiceModuleAttributes


class ReporterServiceModuleAttributes(ScheduleBaseAttributes, ServiceModuleAttributes, metaclass=IterableMeta):
	__config_class__ = "Reporter Service Module"
	host = Attribute('Host')
	log_delivery = Attribute('Log Delivery')
	max_running_reports = Attribute('Max Running Reports', min_version='17.1')
	render_options = Attribute('Render Options', min_version='21.4')
	report_execution_timeout = Attribute('Report Execution Timeout', min_version='17.1')
	report_max_source_record_count = Attribute('Report Max Source Record Count', min_version='17.1')
	smtp_credentials = Attribute('SMTP Credentials')
	secure = Attribute('Secure')
	sender = Attribute('Sender')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_work_base import ClientWorkBaseAttributes
from pytpp.attributes.x509_certificate_base import X509CertificateBaseAttributes


class ClientCertificateWorkAttributes(ClientWorkBaseAttributes, X509CertificateBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Certificate Work"
	certificate_container = Attribute('Certificate Container')
	naming_pattern = Attribute('Naming Pattern')
	transfer_allowed = Attribute('Transfer Allowed', min_version='18.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.metadata_base import MetadataBaseAttributes


class MetadataListAttributes(MetadataBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Metadata List"
	single = Attribute('Single')



from typing import Union
from pytpp.properties.oauth import Scope
from pytpp.api.websdk.websdk import WebSDK
from packaging.version import Version, parse


class Authenticate:
    """
    Authenticates to TPP WebSDK.
    """
    def __init__(self, host: str, username: str = None, password: str = None, application_id: str = None,
                 scope: Union[Scope, str] = None, websdk_token: str = None, proxies: dict = None,
                 certificate_path: str = None, key_file_path: str = None, verify_ssl: bool = False, **kwargs):
        """
        For WebSDK, either an OAuth bearer token can be obtained, which requires both an Application ID and scope
        to be supplied, or the X-Venafi-API-Key can be obtained, which has been deprecated since TPP version 20.1.

        Using the OAuth authentication method requires an API Application Integration to be created via Aperture
        that defines the maximum possible scope and the users/groups that are authorized to use that scope.

        Args:
            host: Hostname or IP Address of TPP
            username: Username
            password: Password
            application_id: Application ID of the OAuth API Application Integration. Must supply ``scope``.
            scope: Scope of the OAuth API Application Integration to be used. Must supply ``application_id``.
            websdk_token: OAuth Access Bearer Token.
            version: Version of the TPP server.
            proxies: An OrderedDict used by the python Requests library.
            certificate_path: Absolute path to the public certificate file.
            key_file_path: Absolute path to the private key file.
            verify_ssl: If ``True``, verify the SSL certificate of the target endpoints.
        """
        self.websdk = WebSDK(host=host, username=username, password=password, token=websdk_token,
                             application_id=application_id, scope=scope, proxies=proxies,
                             certificate_path=certificate_path, key_file_path=key_file_path,
                             verify_ssl=verify_ssl)
        tpp_version = parse(self.websdk.SystemStatus.Version.get().version)
        self._tpp_version = tpp_version
        self.websdk._session.tpp_version = Version(f'{tpp_version.major}.{tpp_version.minor}')
        self._host = host
        self._username = username
        self._password = password
        self._certificate_path = certificate_path
        self._key_file_path = key_file_path
        self._application_id = application_id
        self._scope = scope
        self._proxies = proxies
        self._verify_ssl = verify_ssl

    @property
    def host(self):
        return self._host

    @property
    def username(self):
        return self._username

    @property
    def password(self):
        return self._password

    def re_authenticate(self, host: str = None, username: str = None, password: str = None,
                        application_id: str = None, scope: str = None, proxies: dict = None,
                        certificate_path: str = None, key_file_path: str = None,
                        verify_ssl: bool = None, **kwargs):
        """
        Performs a re-authentication using the same parameters used to authorize initially.

        Args:
            host: Hostname or IP Address
            username: Username
            password: Password
            application_id: Application ID applicable to OAuth. Must supply ``scope``.
            scope: Scope within the Application. Must supply ``application_id``.
            proxies: An OrderedDict used by the python Requests library.
            certificate_path: Absolute path to the public certificate file.
            key_file_path: Absolute path to the private key file.
            verify_ssl: If ``True``, verify the SSL certificate of the target endpoints.
        """
        if any([host, username, password, application_id, scope, certificate_path, key_file_path, proxies]):
            self.__init__(
                host=host or self._host,
                username=username or self._username,
                password=password or self._password,
                application_id=application_id or self._application_id,
                scope=scope or self._scope,
                version=str(self._tpp_version),
                proxies=self._proxies,
                certificate_path=certificate_path or self._certificate_path,
                key_file_path=key_file_path or self._key_file_path,
                verify_ssl=verify_ssl if verify_ssl is not None else self._verify_ssl
            )
        else:
            self.websdk.re_authenticate()



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.username_password_credential import UsernamePasswordCredentialAttributes


class CyberArkUsernamePasswordCredentialAttributes(UsernamePasswordCredentialAttributes, metaclass=IterableMeta):
	__config_class__ = "CyberArk Username Password Credential"
	account_name = Attribute('Account Name', min_version='17.2')
	application_id = Attribute('Application ID', min_version='17.2')
	folder = Attribute('Folder', min_version='17.2')
	safe = Attribute('Safe', min_version='17.2')
	username = Attribute('Username', min_version='17.2')



from pytpp.attributes._helper import IterableMeta, Attribute


class LegacyKeyBaseAttributes(metaclass=IterableMeta):
	__config_class__ = "Legacy Key Base"
	algorithm = Attribute('Algorithm')
	approver = Attribute('Approver')
	key_bit_strength = Attribute('Key Bit Strength')
	key_vault_id = Attribute('Key Vault Id')
	protection_key = Attribute('Protection Key')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class CertificateTrustBundleAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Certificate Trust Bundle"
	excluded_certificates = Attribute('Excluded Certificates')
	included_certificates = Attribute('Included Certificates')
	is_match = Attribute('Is Match')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.report_base import ReportBaseAttributes


class ReportEntitlementAttributes(ReportBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Report:Entitlement"
	options = Attribute('Options')
	policydn = Attribute('PolicyDN', min_version='16.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class PersonAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Person"
	full_name = Attribute('Full Name')
	given_name = Attribute('Given Name')
	group_membership = Attribute('Group Membership')
	internet_email_address = Attribute('Internet EMail Address')
	language = Attribute('Language')
	surname = Attribute('Surname')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.discovery_statistics import DiscoveryStatisticsAttributes
from pytpp.attributes.service_module import ServiceModuleAttributes


class DiscoveryManagerAttributes(DiscoveryStatisticsAttributes, ServiceModuleAttributes, metaclass=IterableMeta):
	__config_class__ = "Discovery Manager"
	connection_timeout = Attribute('Connection Timeout')
	delay = Attribute('Delay')
	load_percentage = Attribute('Load Percentage')
	max_work_units = Attribute('Max Work Units')
	maximum_threads = Attribute('Maximum Threads')
	minimum_threads = Attribute('Minimum Threads')
	placement_disabled = Attribute('Placement Disabled', min_version='16.2')
	timeout = Attribute('Timeout')
	window_days_of_week = Attribute('Window Days of Week')
	window_end = Attribute('Window End')
	window_start = Attribute('Window Start')



from dataclasses import dataclass


@dataclass
class Result:
    code: int
    config_result: str


@dataclass
class AttributeDefinition:
    name: str
    syntax: str


@dataclass
class ClassDefinition:
    containment_names: list
    containment_sub_names: list
    mandatory_names: list
    name: str
    naming_names: list
    optional_names: list
    super_class_names: list



class MetadataFieldType:
    text_string = 1
    list = 2
    date_time = 4
    identity = 5


class MetadataItem:
    allowed_characters = 'AllowedCharacters'
    allowed_values = 'AllowedValues'
    category = 'Category'
    date_only = 'DateOnly'
    default_values = 'DefaultValues'
    display_after = 'DisplayAfter'
    dn = 'DN'
    error_message = 'ErrorMessage'
    guid = 'Guid'
    help = 'Help'
    localization_table = 'LocalizationTable'
    localized_help = 'LocalizedHelp'
    localized_label = 'LocalizedLabel'
    localized_set = 'LocalizedSet'
    mandatory = 'Mandatory'
    name = 'Name'
    mask = 'Mask'
    maximum_length = 'MaximumLength'
    minimum_length = 'MinimumLength'
    policyable = 'Policyable'
    regular_expression = 'RegularExpression'
    render_hidden = 'RenderHidden'
    render_read_only = 'RenderReadOnly'
    single = 'Single'
    time_only = 'TimeOnly'
    type = 'Type'


from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.driver_base import DriverBaseAttributes


class CredentialDriverBaseAttributes(DriverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Credential Driver Base"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_work_base import ClientWorkBaseAttributes


class CertificateProvisioningWorkAttributes(ClientWorkBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Certificate Provisioning Work"
	days_of_month = Attribute('Days Of Month', min_version='15.3')
	days_of_week = Attribute('Days Of Week', min_version='15.3')
	interval = Attribute('Interval', min_version='15.3')
	log_threshold = Attribute('Log Threshold', min_version='15.3')
	schedule_type = Attribute('Schedule Type', min_version='15.3')
	start_time = Attribute('Start Time', min_version='15.3')



from pytpp.properties.resultcodes import ResultCodes
from pytpp.properties.response_objects.dataclasses import credential


class Credential:
    @staticmethod
    def Result(code: int):
        return credential.Result(
            code=code,
            credential_result=ResultCodes.Credential.get(code, 'Unknown'),
        )

    @staticmethod
    def CredentialInfo(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return credential.CredentialInfo(
            class_name=response_object.get('ClassName'),
            full_name=response_object.get('FullName'),
        )

    @staticmethod
    def NameTypeValue(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return credential.NameTypeValue(
            name=response_object.get('Name'),
            type=response_object.get('Type'),
            value=response_object.get('Value'),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class ACMEAccountAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "ACME Account"
	account_key_hash = Attribute('Account Key Hash', min_version='17.2')
	certificate_location_dn = Attribute('Certificate Location DN', min_version='17.2')
	internet_email_address = Attribute('Internet EMail Address', min_version='17.2')
	json_web_key = Attribute('JSON Web Key', min_version='21.2')
	public_key_vault_id = Attribute('Public Key Vault Id', min_version='17.2')
	status = Attribute('Status', min_version='17.2')



from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.stats import Stats


class _Stats:
    def __init__(self, api_obj):
        self.GetCounters = self._GetCounters(api_obj=api_obj)
        self.Query = self._Query(api_obj=api_obj)
    
    class _GetCounters(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='Stats/GetCounters')

        def post(self):
            
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def counters(self):
                    return [Stats.Counter(counter) for counter in self._from_json(key='Counters')]

                @property
                @api_response_property()
                def error(self) -> str:
                    return self._from_json(key='Error')

            return _Response(response=self._post(data={}))

    class _Query(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='Stats/Query')

        def post(self, stats_type: str, tier: int, max_points: int, group_by_a: bool = None, group_by_b: bool = None,
                 group_by_c: bool = None, filter_a: str = None, filter_b: str = None, filter_c: str = None):
            body = {
                'StatsType': stats_type,
                'Tier': tier,
                'MaxPoints': max_points,
                'GroupByA': group_by_a,
                'GroupByB': group_by_b,
                'GroupByC': group_by_c,
                'FilterA': filter_a,
                'FilterB': filter_b,
                'FilterC': filter_c
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def results(self):
                    return [Stats.Result(result) for result in self._from_json(key='Results')]

            return _Response(response=self._post(data=body))



from pytpp.properties.response_objects.dataclasses import oauth


class OAuth:
    @staticmethod
    def Permissions(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return oauth.Permissions(
            delete=response_object.get('delete'),
            discover=response_object.get('discover'),
            manage=response_object.get('manage'),
            read=response_object.get('read'),
            revoke=response_object.get('revoke'),
        )



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.top import TopAttributes


class RemoteAccessApplicationContainerAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Remote Access Application Container"



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.top import TopAttributes


class ReportContainerAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Report Container"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class TealeafPCAAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Tealeaf PCA"
	file_validation_disabled = Attribute('File Validation Disabled')
	install_path = Attribute('Install Path')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.code_signing_environment_template_base import CodeSigningEnvironmentTemplateBaseAttributes


class CodeSigningCertificateEnvironmentTemplateAttributes(CodeSigningEnvironmentTemplateBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing Certificate Environment Template"
	certificate_authority = Attribute('Certificate Authority', min_version='20.1')
	certificate_container_dn = Attribute('Certificate Container DN', min_version='20.1')
	city = Attribute('City', min_version='20.1')
	country = Attribute('Country', min_version='20.1')
	key_algorithm = Attribute('Key Algorithm', min_version='20.1')
	key_storage_location = Attribute('Key Storage Location', min_version='20.1')
	organization = Attribute('Organization', min_version='20.1')
	organizational_unit = Attribute('Organizational Unit', min_version='20.1')
	state = Attribute('State', min_version='20.1')
	x509_subject = Attribute('X509 Subject', min_version='20.3')
	x509_subjectaltname_rfc822 = Attribute('X509 SubjectAltName RFC822', min_version='20.1')



from typing import List
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.secret_store import SecretStore


class _X509CertificateStore:
    def __init__(self, api_obj):
        self.Add = self._Add(api_obj=api_obj)
        self.Lookup = self._Lookup(api_obj=api_obj)
        self.LookupExpiring = self._LookupExpiring(api_obj=api_obj)
        self.Remove = self._Remove(api_obj=api_obj)
        self.Retrieve = self._Retrieve(api_obj=api_obj)

    class _Add(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/X509CertificateStore/Add')

        def post(self, owner_dn: str, certificate_collection_strings: list = None, certificate_string: str = None,
                 protection_key: str = None, typed_name_values: list = None):
            body = {
                'CertificateCollectionStrings': certificate_collection_strings,
                'CertificateString': certificate_string,
                'OwnerDN': owner_dn,
                'ProtectionKey': protection_key,
                'TypedNameValues': typed_name_values
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def leaf_existed(self) -> bool:
                    return self.api_response('LeafExisted')

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self.api_response('Result'))

                @property
                @api_response_property()
                def vault_id(self) -> int:
                    return self.api_response('VaultId')

            return _Response(response=self._post(data=body))

    class _Lookup(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/X509CertificateStore/Lookup')

        def post(self, certificate_string: str = None, name: str = None, owner_dn: str = None, value: str = None):
            body = {
                'CertificateString': certificate_string,
                'Name': name,
                'OwnerDN': owner_dn,
                'Value': value
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def vault_id(self) -> int:
                    return self.api_response('VaultId')

                @property
                @api_response_property()
                def vault_ids(self) -> List[int]:
                    return self.api_response('VaultIds')

                @property
                @api_response_property()
                def certificate_collection_strings(self) -> List[str]:
                    return self.api_response('CertificateCollectionStrings')

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self.api_response('Result'))

            return _Response(response=self._post(data=body))

    class _LookupExpiring(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/X509CertificateStore/LookupExpiring')

        def post(self, days_to_expiration: int, owner_dn: str):
            body = {
                'DaysToExpiration': days_to_expiration,
                'OwnerDN': owner_dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def vault_ids(self) -> List[int]:
                    return self.api_response('VaultIds')

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self.api_response('Result'))

            return _Response(response=self._post(data=body))

    class _Remove(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/X509CertificateStore/Remove')

        def post(self, owner_dn: str, certificate: str = None, vault_id: int = None):
            body = {
                'Certificate': certificate,
                'OwnerDN': owner_dn,
                'VaultId': vault_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self.api_response('Result'))

            return _Response(response=self._post(data=body))

    class _Retrieve(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/X509CertificateStore/Retrieve')

        def post(self, vault_id: int):
            body = {
                'VaultId': vault_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def certificate_string(self) -> str:
                    return self.api_response('CertificateString')

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self.api_response('Result'))

                @property
                @api_response_property()
                def typed_name_values(self):
                    return SecretStore.TypedNameValues(self.api_response('TypedNameValues'))

            return _Response(response=self._post(data=body))



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.top import TopAttributes


class ClientGroupContainerAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Group Container"



from pytpp.attributes._helper import IterableMeta, Attribute


class TopAttributes(metaclass=IterableMeta):
	__config_class__ = "Top"
	contact = Attribute('Contact')
	created_by = Attribute('Created By', min_version='15.1')
	description = Attribute('Description')
	disabled = Attribute('Disabled')
	escalation_contact = Attribute('Escalation Contact')
	guid = Attribute('GUID')
	managed_by = Attribute('Managed By')
	metadata = Attribute('Metadata')
	reference = Attribute('Reference')
	workflow = Attribute('Workflow')
	workflow_block = Attribute('Workflow Block')



from typing import Union
from pytpp.tools.vtypes import Config, Identity
from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.properties.response_objects.permissions import Permissions as PermResponseObj


@feature('Permissions')
class Permissions(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)

    def _get_obj_and_identity(self, obj: 'Union[Config.Object, str]', identity: 'Union[Identity.Identity, str]'):
        if isinstance(obj, str):
            obj = self._get_config_object(obj)
        if isinstance(identity, str):
            identity = self._get_identity_object(identity)
        return obj, identity

    def delete(self, obj: 'Union[Config.Object, str]', identity: 'Union[Identity.Identity, str]'):
        """
        Deletes all explicit permissions granted to a user or group on the ``obj``. All implicit permissions,
        i.e. those that are inherited from group memberships and parent folders, are unaffected.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.
            identity: :ref:`identity_object` or :ref:`prefixed_name` of the user or group.
        """
        obj, identity = self._get_obj_and_identity(obj=obj, identity=identity)
        current_permissions = self.get_explicit(obj=obj, identity=identity)
        if not current_permissions:
            return

        if '+' in identity.prefix:
            ptype, pname = identity.prefix.split('+', 1)
            api = self._api.websdk.Permissions.Object.Guid(obj.guid).Ptype(ptype).Pname(pname).Principal(identity.universal)
        else:
            api = self._api.websdk.Permissions.Object.Guid(obj.guid).Ptype().Principal(identity.universal)

        result = api.delete()
        result.assert_valid_response()

    def get_effective(self, obj: 'Union[Config.Object, str]', identity: 'Union[Identity.Identity, str]'):
        """
        Returns the *effective* permissions of a user or group on the ``obj``. Effective permissions are the
        permissions that are take effect when the user authenticates to TPP. All Master Admin, implicit, and
        explicit permissions are taken into account to evaluate the final effective permissions of a user or group.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.
            identity: :ref:`identity_object` or :ref:`prefixed_name` of the user or group.

        Returns:
            :py:class:`~.dataclasses.permissions.Permissions`: Effective permissions granted to the ``identity``.
        """
        obj, identity = self._get_obj_and_identity(obj=obj, identity=identity)
        if '+' in identity.prefix:
            ptype, pname = identity.prefix.split('+', 1)
            api = self._api.websdk.Permissions.Object.Guid(obj.guid).Ptype(ptype).Pname(pname).Principal(identity.universal)
        else:
            api = self._api.websdk.Permissions.Object.Guid(obj.guid).Ptype().Principal(identity.universal)

        result = api.Effective.get()
        return result.effective_permissions if result.is_valid_response() else PermResponseObj.Permissions({})

    def get_explicit(self, obj: 'Union[Config.Object, str]', identity: 'Union[Identity.Identity, str]'):
        """
        Returns the `explicit` permissions of a user or group on the ``obj``. Explicit permissions are the
        permissions that are `explicitly` granted to a user or group on a particular object. A user or group may
        have permissions to the object via `implicit` permissions, which are permissions inherited from other
        folders and group memberships. Implicit permissions are ignored. To get implicit permissions, use
        :meth:`~get_implicit`.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.
            identity: :ref:`identity_object` or :ref:`prefixed_name` of the user or group.

        Returns:
            :py:class:`~.dataclasses.permissions.Permissions`: Explicit permissions granted to the ``identity``.
        """
        obj, identity = self._get_obj_and_identity(obj=obj, identity=identity)
        if '+' in identity.prefix:
            ptype, pname = identity.prefix.split('+', 1)
            api = self._api.websdk.Permissions.Object.Guid(obj.guid).Ptype(ptype).Pname(pname).Principal(identity.universal)
        else:
            api = self._api.websdk.Permissions.Object.Guid(obj.guid).Ptype().Principal(identity.universal)

        result = api.get()
        return result.explicit_permissions if result.is_valid_response() else PermResponseObj.Permissions({})

    def get_implicit(self, obj: 'Union[Config.Object, str]', identity: 'Union[Identity.Identity, str]'):
        """
        Returns the `implicit` permissions of a user or group on the ``obj``. Implicit permissions are permissions
        inherited from other folders and group memberships. To get explicit permissions, use :meth:`get_explicit`.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.
            identity: :ref:`identity_object` or :ref:`prefixed_name` of the user or group.

        Returns:
            :py:class:`~.dataclasses.permissions.Permissions`: Implicit permissions granted to the ``identity``.
        """
        obj, identity = self._get_obj_and_identity(obj=obj, identity=identity)
        if '+' in identity.prefix:
            ptype, pname = identity.prefix.split('+', 1)
            api = self._api.websdk.Permissions.Object.Guid(obj.guid).Ptype(ptype).Pname(pname).Principal(identity.universal)
        else:
            api = self._api.websdk.Permissions.Object.Guid(obj.guid).Ptype().Principal(identity.universal)

        result = api.get()
        return result.implicit_permissions if result.is_valid_response() else PermResponseObj.Permissions({})

    def list_identities(self, obj: 'Union[Config.Object, str]'):
        """
        Returns a list of Identity objects that have `explicit` permissions to the object. Explicit permissions are the
        permissions that are `explicitly` granted to a user or group on a particular object. A user or group may
        have permissions to the object via `implicit` permissions, which are permissions inherited from other
        folders and group memberships. Implicit permissions are ignored.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.

        Returns:
            List of :ref:`identity_object`.
        """
        obj = self._get_config_object(obj)
        principals = self._api.websdk.Permissions.Object.Guid(obj.guid).get().principals

        principals = [
            self._api.websdk.Identity.Validate.post(identity={'PrefixedUniversal': principal}).identity
            for principal in principals
        ]

        return principals

    def update(self, obj: 'Union[Config.Object, str]', identity: 'Union[Identity.Identity, str]', is_associate_allowed: bool = None,
               is_create_allowed: bool = None, is_delete_allowed: bool = None, is_manage_permissions_allowed: bool = None,
               is_policy_write_allowed: bool = None, is_private_key_read_allowed: bool = None, is_private_key_write_allowed: bool = None,
               is_read_allowed: bool = None, is_rename_allowed: bool = None, is_revoke_allowed: bool = None, is_view_allowed: bool = None,
               is_write_allowed: bool = None):
        """
        Grants the specified permissions to a user or group identity. If any arguments are not specified as
        ``True`` or ``False`` then that value will default to their existing permissions or ``False``.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.
            identity: :ref:`identity_object` or :ref:`prefixed_name` of the user or group.
            is_associate_allowed: Allows associating/dissociating applications to certificates and pushing certificates
                to the associated applications.
            is_create_allowed: Allows creating subordinate objects to the ``obj``. Also grants View permission.
            is_delete_allowed: Allows deleting subordinate objects to the ``obj``.
            is_manage_permissions_allowed: Allows modification to others' permissions to ``obj`` and its
                subordinate objects.
            is_policy_write_allowed: Allows modification to policy values on folder. Requires View permission. Also grants
                Read and Write permissions.
            is_private_key_read_allowed: Allows download of private keys.
            is_private_key_write_allowed: Allows upload of private keys.
            is_read_allowed: Allows ability to read values on subordinate objects to ``obj``.
            is_rename_allowed: Allows ability to rename and move subordinate objects to ``obj``. Requires Rename
                permission to the destination location.
            is_revoke_allowed: Allows ability to invalidate a certificate. Requires Write permission to the certificate
                object.
            is_view_allowed: Allows ability to view the name of all subordinate objects to ``obj``.
            is_write_allowed: Allows editing of subordinate objects to ``obj``.
        """
        obj, identity = self._get_obj_and_identity(obj=obj, identity=identity)
        if '+' in identity.prefix:
            ptype, pname = identity.prefix.split('+', 1)
            api = self._api.websdk.Permissions.Object.Guid(obj.guid).Ptype(ptype).Pname(pname).Principal(identity.universal)
        else:
            api = self._api.websdk.Permissions.Object.Guid(obj.guid).Ptype().Principal(identity.universal)

        current_permissions = self.get_explicit(obj=obj, identity=identity)

        if bool([y for x, y in current_permissions.__dict__.items() if not x.startswith('_') and y is not None]):
            method = api.put
        else:
            method = api.post

        new_permissions = {
            k: v if v is not None else getattr(current_permissions, k) for k, v in dict(
                is_associate_allowed=is_associate_allowed,
                is_create_allowed=is_create_allowed,
                is_delete_allowed=is_delete_allowed,
                is_manage_permissions_allowed=is_manage_permissions_allowed,
                is_policy_write_allowed=is_policy_write_allowed,
                is_private_key_read_allowed=is_private_key_read_allowed,
                is_private_key_write_allowed=is_private_key_write_allowed,
                is_read_allowed=is_read_allowed,
                is_rename_allowed=is_rename_allowed,
                is_revoke_allowed=is_revoke_allowed,
                is_view_allowed=is_view_allowed,
                is_write_allowed=is_write_allowed,
            ).items()
        }

        result = method(**new_permissions)
        result.assert_valid_response()



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class OpenTrustPKICAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "OpenTrust PKI CA"
	connector_type = Attribute('Connector Type')
	fields = Attribute('Fields')
	retrieval_period = Attribute('Retrieval Period')
	web_service_url = Attribute('Web Service URL')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.branch_base import BranchBaseAttributes


class RemoteAccessRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Remote Access Root"
	approved_issuer = Attribute('Approved Issuer', min_version='19.2')
	authentication_methods = Attribute('Authentication Methods', min_version='19.2')
	certificate_auth_lookup_field = Attribute('Certificate Auth Lookup Field', min_version='19.2')
	certificate_auth_source_field = Attribute('Certificate Auth Source Field', min_version='19.2')
	certificate_auth_source_regex = Attribute('Certificate Auth Source Regex', min_version='19.2')
	certificate_pinning_options = Attribute('Certificate Pinning Options', min_version='19.2')
	certificate_pinning_work_dn = Attribute('Certificate Pinning Work DN', min_version='19.2')
	expiration = Attribute('Expiration', min_version='20.1')
	grant_validity = Attribute('Grant Validity', min_version='19.2')
	mapping_table = Attribute('Mapping Table', min_version='20.1')
	refresh_api_enabled = Attribute('Refresh API Enabled', min_version='20.1')
	refresh_token_enabled = Attribute('Refresh Token Enabled', min_version='19.2')
	session_pool_entries = Attribute('Session Pool Entries', min_version='19.2')
	session_pool_max_age = Attribute('Session Pool Max Age', min_version='19.2')
	session_rights_refresh_age = Attribute('Session Rights Refresh Age', min_version='19.2')
	token_validity = Attribute('Token Validity', min_version='19.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_group_base import ClientGroupBaseAttributes


class ClientGroupAttributes(ClientGroupBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Group"
	agent_type = Attribute('Agent Type', min_version='15.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.service_module import ServiceModuleAttributes


class SSHManagerAttributes(ServiceModuleAttributes, metaclass=IterableMeta):
	__config_class__ = "SSH Manager"
	escalation_notice_interval = Attribute('Escalation Notice Interval')
	escalation_notice_start = Attribute('Escalation Notice Start')
	expiration_notice_interval = Attribute('Expiration Notice Interval')
	expiration_notice_start = Attribute('Expiration Notice Start')
	renewal_window = Attribute('Renewal Window')
	renewal_window_event = Attribute('Renewal Window Event')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.flow_action_base import FlowActionBaseAttributes


class CreateKeyActionAttributes(FlowActionBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Create Key Action"
	algorithm = Attribute('Algorithm', min_version='20.2')
	key_storage_location = Attribute('Key Storage Location', min_version='20.2')
	vault_owner = Attribute('Vault Owner', min_version='20.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.report_base import ReportBaseAttributes
from pytpp.attributes.report_filter_base import ReportFilterBaseAttributes


class ReportTrustAttributes(ReportBaseAttributes, ReportFilterBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Report:Trust"
	discoverydn = Attribute('DiscoveryDN')
	grouping = Attribute('Grouping')
	options = Attribute('Options')
	trustedca = Attribute('TrustedCA')
	untrustedca = Attribute('UntrustedCA')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class GSKAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "GSK"
	backup_store = Attribute('Backup Store')
	certificate_label = Attribute('Certificate Label')
	create_store = Attribute('Create Store')
	default_cert = Attribute('Default Cert')
	disable_ssh_history = Attribute('Disable SSH History', min_version='19.1')
	file_validation_disabled = Attribute('File Validation Disabled', min_version='15.3')
	fips_key = Attribute('Fips Key')
	hide_command_line_passwords = Attribute('Hide Command Line Passwords', min_version='19.4')
	java_home_path = Attribute('Java Home Path')
	key_store = Attribute('Key Store')
	key_store_credential = Attribute('Key Store Credential')
	key_store_validation_disabled = Attribute('Key Store Validation Disabled')
	network_validation_disabled = Attribute('Network Validation Disabled')
	password_expire_days = Attribute('Password Expire Days')
	recycle_alias = Attribute('Recycle Alias')
	refresh_security = Attribute('Refresh Security')
	replace_store = Attribute('Replace Store')
	stash_password = Attribute('Stash Password')
	store_type = Attribute('Store Type')
	temp_certificate_label = Attribute('Temp Certificate Label')
	utility_path = Attribute('Utility Path')
	version = Attribute('Version')



from typing import List
from datetime import datetime
from dataclasses import dataclass


@dataclass
class Response:
    success: bool
    error_code: int
    error_message: str


@dataclass
class ConnectionResult:
    device_guid: str
    error: str
    success: bool


@dataclass
class DeviceData:
    dn: str
    device_guid: str
    host_name: str
    is_compliant: bool
    type: str


@dataclass
class KeyData:
    active_from: datetime
    algorithm: str
    allowed_source_restriction: List[str]
    approver: List[str]
    comment: str
    denied_source_restriction: List[str]
    device_guid: str
    filepath: str
    forced_command: str
    format: str
    is_encrypted: bool
    key_id: int
    keysetid: str
    last_used: datetime
    length: int
    notes: str
    options: List[str]
    process_error: str
    process_status: str
    reason: str
    rotation_stage: int
    type: str
    username: str
    violation_status: List[int]


@dataclass
class KeySetData:
    access: str
    algorithm: str
    fingerprint_md5: str
    fingerprint_sha256: str
    keysetid: str
    last_rotation_date: datetime
    last_used: datetime
    length: int
    private_keys: 'List[KeyData]'
    process_error: str
    process_status: str
    public_keys: 'List[KeyData]'
    rotation_stage: int
    type: str
    violation_status: list


@dataclass
class KeyUsageData:
    alert: int
    authorized_key_id: int
    client_name: str
    fingerprint: str
    key_usage_id: int
    keyset_id: str
    last_used: datetime
    private_key_id: int
    server_account: str
    server_name: str



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_work_base import ClientWorkBaseAttributes


class ClientAgentDevicePlacementWorkAttributes(ClientWorkBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Agent Device Placement Work"
	device_object_location = Attribute('Device Object Location')
	ssh_device_object_location = Attribute('SSH Device Object Location', min_version='15.2')
	ssl_device_object_location = Attribute('SSL Device Object Location', min_version='15.2')



class CertificateAttributes:
    country = 'C'
    certificate_type = 'CertificateType'
    common_name = 'CN'
    issuer = 'Issuer'
    is_self_signed = 'IsSelfSigned'
    is_wild_card = 'IsWildCard'
    key_algorithm = 'KeyAlgorithm'
    key_size = 'KeySize'
    key_size_generator = 'KeySizeGenerator'
    key_size_less = 'KeySizeLess'
    locality = 'L'
    organization = 'O'
    organization_unit = 'OU'
    state = 'S'
    san_dns = 'SAN-DNS'
    san_email = 'SAN-Email'
    san_ip = 'SAN-IP'
    san_upn = 'SAN-UPN'
    san_uri = 'SAN-URI'
    serial = 'Serial'
    signature_algorithm = 'SignatureAlgorithm'
    thumbprint = 'Thumbprint'
    valid_from = 'ValidFrom'
    valid_to = 'ValidTo'
    valid_to_greater = 'ValidToGreater'
    valid_to_less = 'ValidToLess'


class CertificateStatus:
    chain_validation_failure = 'ChainValidationFailure'
    created_on = 'CreatedOn'
    created_on_greater = 'CreatedOnGreater'
    created_on_less = 'CreatedOnLess'
    disabled = 'Disabled'
    in_error = 'InError'
    management_type = 'ManagementType'
    name = 'Name'
    network_validation_disabled = 'NetworkValidationDisabled'
    parent_dn = 'ParentDn'
    parent_dn_recursive = 'ParentDnRecursive'
    pending_workflow = 'PendingWorkflow'
    ssl_tls_protocol = 'SslTlsProtocol'
    stage = 'Stage'
    stage_greater = 'StageGreater'
    stage_less = 'StageLess'
    tls_validation_failure = 'TlsValidationFailure'
    validation_disabled = 'ValidationDisabled'
    validation_state = 'ValidationState'


class OptionalFields:
    issuer = 'Issuer'
    key_algorithm = 'KeyAlgorithm'
    key_size = 'KeySize'
    subject = 'Subject'



from pytpp import Authenticate, Scope

scope = Scope()
scope.certificate(approve=True)
scope.security(delete=True)

api = Authenticate('mctpp1.venspi.eng.venafi.com', 'admin', 'newPassw0rd!')

print('done')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.service_module import ServiceModuleAttributes


class CertificateRevocationAttributes(ServiceModuleAttributes, metaclass=IterableMeta):
	__config_class__ = "Certificate Revocation"
	ca_issuer_monitor_disabled = Attribute('CA Issuer Monitor Disabled', min_version='19.2')
	ocsp_concurrent_connection_limit = Attribute('OCSP Concurrent Connection Limit', min_version='19.2')
	ocsp_concurrent_request_limit = Attribute('OCSP Concurrent Request Limit', min_version='19.2')



from pytpp.properties.response_objects.dataclasses import rights


class Rights:
    @staticmethod
    def Rights(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return rights.Rights(
            checksum=response_object.get('Checksum'),
            is_container=response_object.get('IsContainer'),
            is_group=response_object.get('IsGroup'),
            object=response_object.get('Object'),
            principal=response_object.get('Principal'),
            rights=response_object.get('Rights'),
            sub_system=response_object.get('SubSystem'),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.report_base import ReportBaseAttributes
from pytpp.attributes.report_filter_base import ReportFilterBaseAttributes


class ReportCustomReportAttributes(ReportBaseAttributes, ReportFilterBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Report:Custom Report"
	custom_report_data_type = Attribute('Custom Report: Data Type', min_version='15.3')
	custom_report_entity_type = Attribute('Custom Report: Entity Type', min_version='19.2')
	custom_report_filter = Attribute('Custom Report: Filter', min_version='15.3')
	custom_report_order_by = Attribute('Custom Report: Order By', min_version='15.3')
	custom_report_owner = Attribute('Custom Report: Owner', min_version='16.1')
	custom_report_personalized = Attribute('Custom Report: Personalized', min_version='15.3')
	custom_report_selected_field = Attribute('Custom Report: Selected Field', min_version='15.3')
	custom_report_sort_descending = Attribute('Custom Report: Sort Descending', min_version='15.3')
	error_message = Attribute('Error Message', min_version='15.3')
	options = Attribute('Options', min_version='15.3')
	status = Attribute('Status', min_version='15.3')
	summary = Attribute('Summary', min_version='15.3')



from dataclasses import dataclass


@dataclass
class Permissions:
    is_associate_allowed: bool
    is_create_allowed: bool
    is_delete_allowed: bool
    is_manage_permissions_allowed: bool
    is_policy_write_allowed: bool
    is_private_key_read_allowed: bool
    is_private_key_write_allowed: bool
    is_read_allowed: bool
    is_rename_allowed: bool
    is_revoke_allowed: bool
    is_view_allowed: bool
    is_write_allowed: bool



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.flow_action_base import FlowActionBaseAttributes


class FlowActionAddTeamMemberAttributes(FlowActionBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Flow Action Add Team Member"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes
from pytpp.attributes.proxy import ProxyAttributes


class zOSCAAttributes(HTTPCABaseAttributes, ProxyAttributes, metaclass=IterableMeta):
	__config_class__ = "zOS CA"
	client_port = Attribute('Client Port')
	client_secure = Attribute('Client Secure')
	domain = Attribute('Domain')
	secure = Attribute('Secure')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class ClientWorkBaseAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Work Base"
	work_guid = Attribute('Work Guid')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class LayoutRuleBaseAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Layout Rule Base"
	rule = Attribute('Rule', min_version='19.1')
	vault_id = Attribute('Vault Id')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class ApertureConfigurationAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Aperture Configuration"
	applies_to = Attribute('Applies To')
	feedback_url = Attribute('Feedback Url')
	help_text = Attribute('Help Text')
	help_url = Attribute('Help Url')
	lost_and_found = Attribute('Lost and Found')
	timeout = Attribute('Timeout')



from typing import Union
from datetime import datetime, timezone, timedelta
from dateutil.parser import parse
import isodate


def from_date_string(date_string: str, duration_format: bool = False) -> Union[None, datetime]:
    if not date_string:
        return None

    if duration_format:
        return isodate.parse_duration(datestring=date_string)

    if 'Date' in date_string:
        date_as_num = ''.join([c for c in date_string if c.isnumeric() or c in ('+', '-')])
        if '+' in date_as_num:
            dt, tz = date_as_num.split('+')
            tz = timezone(timedelta(hours=int(tz) / 100))
        elif '-' in date_as_num:
            dt, tz = date_as_num.split('-')
            tz = timezone(timedelta(hours=int(tz)/100))
        else:
            dt, tz = date_as_num, None
        date_as_float = float(int(dt) / 1000)
        return datetime.fromtimestamp(date_as_float, tz=tz)

    else:
        return parse(date_string)



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.branch_base import BranchBaseAttributes


class DiscoveryRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Discovery Root"
	address_range = Attribute('Address Range')
	dsn = Attribute('DSN')
	driver_name = Attribute('Driver Name')
	protection_key = Attribute('Protection Key')
	timeout = Attribute('Timeout')



from dataclasses import dataclass
from typing import Optional


@dataclass
class Result:
    code: int
    config_result: str


@dataclass
class NameValues:
    name: str
    values: list


@dataclass
class Object:
    absolute_guid: str
    dn: str
    guid: str
    config_id: Optional[int]
    name: str
    parent: str
    revision: Optional[int]
    type_name: str


@dataclass
class Policy:
    attribute_name: str
    guid: str
    property: str
    type_name: str
    value_list: list



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.x509_root_certificate import X509RootCertificateAttributes


class X509IntermediateRootCertificateAttributes(X509RootCertificateAttributes, metaclass=IterableMeta):
	__config_class__ = "X509 Intermediate Root Certificate"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class NetScalerAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "NetScaler"
	certificate_file = Attribute('Certificate File')
	chain_cert = Attribute('Chain Cert')
	file_validation_disabled = Attribute('File Validation Disabled')
	fips_key = Attribute('Fips Key')
	import_only = Attribute('Import Only')
	install_path = Attribute('Install Path', min_version='20.2')
	issuer_certificate_name = Attribute('Issuer Certificate Name', min_version='15.1')
	max_filesize = Attribute('Max Filesize')
	network_validation_disabled = Attribute('Network Validation Disabled')
	private_key_file = Attribute('Private Key File')
	sni_certificate = Attribute('SNI Certificate', min_version='17.1')
	ssl_object_type = Attribute('SSL Object Type', min_version='17.1')
	temp_certificate_label = Attribute('Temp Certificate Label')
	virtual_server_name = Attribute('Virtual Server Name')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.x509_certificate import X509CertificateAttributes


class X509CodeSigningCertificateAttributes(X509CertificateAttributes, metaclass=IterableMeta):
	__config_class__ = "X509 Code Signing Certificate"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.driver_base import DriverBaseAttributes


class AgentModuleHandlerAttributes(DriverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Agent Module Handler"
	version = Attribute('Version')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_trust_store_base import CertificateTrustStoreBaseAttributes


class PEMTrustStoreAttributes(CertificateTrustStoreBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "PEM Trust Store"
	key_store = Attribute('Key Store')



from pytpp.properties.resultcodes import ResultCodes
from pytpp.tools.helpers.date_converter import from_date_string
from pytpp.properties.response_objects.dataclasses import codesign


class CodeSign:
    @staticmethod
    def ResultCode(code: int):
        return codesign.ResultCode(
            code=code,
            codesign_result=ResultCodes.CodeSign.get(code, 'Unknown'),
        )

    @staticmethod
    def Items(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.Items(
            items=response_object.get('Items'),
        )

    @staticmethod
    def InfoValue(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.InfoValue(
            info=response_object.get('Info'),
            value=CodeSign.Items(response_object.get('Value')),
        )

    @staticmethod
    def EnvironmentDetails(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.EnvironmentTemplateDetails(
            info=response_object.get('Info'),
            value=response_object.get('Value'),
            template_values=CodeSign.InfoValue(response_object.get('TemplateValues')),
        )

    @staticmethod
    def RightsKeyValue(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.RightsKeyValue(
            key=response_object.get('key'),
            value=response_object.get('value'),
        )

    @staticmethod
    def CertificateTemplate(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.CertificateTemplate(
            allow_user_key_import=response_object.get('AllowUserKeyImport'),
            certificate_authority_dn=CodeSign.InfoValue(response_object.get('CertificateAuthorityDN')),
            certificate_subject=response_object.get('CertificateSubject'),
            city=CodeSign.InfoValue(response_object.get('City')),
            country=CodeSign.InfoValue(response_object.get('Country')),
            dn=response_object.get('Dn'),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            key_algorithm=CodeSign.InfoValue(response_object.get('KeyAlgorithm')),
            key_storage_location=CodeSign.InfoValue(response_object.get('KeyStorageLocation')),
            key_use_flow_dn=response_object.get('KeyUseFlowDN'),
            object_naming_pattern=response_object.get('ObjectNamingPattern'),
            organization=CodeSign.InfoValue(response_object.get('Organization')),
            organization_unit=CodeSign.InfoValue(response_object.get('OrganizationalUnit')),
            per_user=response_object.get('PerUser'),
            read_only=response_object.get('ReadOnly'),
            san_email=CodeSign.InfoValue(response_object.get('SANEmail')),
            state=CodeSign.InfoValue(response_object.get('State')),
            target_policy_dn=response_object.get('TargetPolicyDN'),
            type=response_object.get('Type'),
            visible_to=CodeSign.Items(response_object.get('VisibleTo')),
        )

    @staticmethod
    def CustomFieldAttributes(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.CustomFieldAttributes(
            field_name=response_object.get('FieldName'),
            values=response_object.get('Values'),
        )

    @staticmethod
    def CertificateEnvironment(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.CertificateEnvironment(
            allow_user_key_import=response_object.get('AllowUserKeyImport'),
            ca_specific_attributes=CodeSign.Items(response_object.get('CASpecificAttributes')),
            certificate_authority_dn=CodeSign.InfoValue(response_object.get('CertificateAuthorityDN')),
            certificate_subject=response_object.get('CertificateSubject'),
            certificate_template=CodeSign.CertificateTemplate(response_object.get('CertificateTemplate')),
            city=CodeSign.InfoValue(response_object.get('City')),
            country=CodeSign.InfoValue(response_object.get('Country')),
            custom_field_attributes=[
                CodeSign.CustomFieldAttributes(cfa) for cfa in response_object.get('CustomFieldAttributes', [])
            ],
            dn=response_object.get('Dn'),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            ip_address_restriction=CodeSign.Items(response_object.get('IPAddressRestriction')),
            key_algorithm=CodeSign.InfoValue(response_object.get('KeyAlgorithm')),
            key_storage_location=CodeSign.InfoValue(response_object.get('KeyStorageLocation')),
            key_time_constraints=CodeSign.Items(response_object.get('KeyTimeConstraints')),
            key_use_flow_dn=response_object.get('KeyUseFlowDN'),
            organization=CodeSign.InfoValue(response_object.get('Organization')),
            organization_unit=CodeSign.InfoValue(response_object.get('OrganizationUnit')),
            per_user=response_object.get('PerUser'),
            san_email=CodeSign.InfoValue(response_object.get('SANEmail')),
            state=CodeSign.InfoValue(response_object.get('State')),
            status=response_object.get('Status'),
            target_store=response_object.get('TargetStore'),
            template_dn=response_object.get('TemplateDN'),
            type=response_object.get('Type'),
        )

    @staticmethod
    def KeyStorageLocations(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.KeyStorageLocations(
            items=response_object.get('Items'),
        )

    @staticmethod
    def Application(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.Application(
            dn=response_object.get('Dn'),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
        )

    @staticmethod
    def ApplicationCollection(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.ApplicationCollection(
            application_dns=CodeSign.Items(response_object.get('ApplicationDNs')),
            dn=response_object.get('Dn'),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
        )

    @staticmethod
    def AppleEnvironment(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.AppleEnvironment(
            allow_user_key_import=response_object.get('AllowUserKeyImport'),
            apple_template=CodeSign.AppleTemplate(response_object.get('AppleTemplate')),
            custom_field_attributes=[
                CodeSign.CustomFieldAttributes(cfa) for cfa in
                response_object.get('CustomFieldAttributes')
            ],
            dirty=response_object.get('Dirty'),
            dn=response_object.get('Dn'),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            ip_address_restriction=CodeSign.Items(response_object.get('IPAddressRestriction')),
            key_time_constraints=CodeSign.Items(response_object.get('KeyTimeConstraints')),
            key_use_flow_dn=response_object.get('KeyUseFlowDN'),
            per_user=response_object.get('PerUser'),
            template_dn=response_object.get('TemplateDN'),
        )

    @staticmethod
    def AppleTemplate(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.AppleTemplate(
            allow_user_key_import=response_object.get('AllowUserKeyImport'),
            cn_pattern=response_object.get('CnPattern'),
            dn=response_object.get('Dn'),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            key_storage_location=CodeSign.InfoValue(response_object.get('KeyStorageLocation')),
            key_time_constraints=CodeSign.Items(response_object.get('KeyTimeConstraints')),
            key_use_flow_dn=response_object.get('KeyUseFlowDN'),
            read_only=response_object.get('ReadOnly'),
            target_policy_dn=response_object.get('TargetPolicyDN'),
            type=response_object.get('Type'),
            visible_to=CodeSign.Items(response_object.get('VisibleTo')),
        )

    @staticmethod
    def CSPTemplate(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.CSPTemplate(
            allow_user_key_import=response_object.get('AllowUserKeyImport'),
            dirty=response_object.get('Dirty'),
            dn=response_object.get('Dn'),
            encryption_key_algorithm=CodeSign.InfoValue(response_object.get('EncryptionKeyAlgorithm')),
            expiration=CodeSign.InfoValue(response_object.get('Expiration')),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            key_container_dn=response_object.get('KeyContainerDN'),
            key_storage_location=CodeSign.InfoValue(response_object.get('KeyStorageLocation')),
            max_uses=CodeSign.InfoValue(response_object.get('MaxUses')),
            object_naming_pattern=response_object.get('ObjectNamingPattern'),
            per_user=response_object.get('PerUser'),
            signing_key_algorithm=CodeSign.InfoValue(response_object.get('SigningKeyAlgorithm')),
            type=response_object.get('Type'),
            visible_to=CodeSign.Items(response_object.get('VisibleTo')),
        )

    @staticmethod
    def CSPEnvironment(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.CSPEnvironment(
            allow_user_key_import=response_object.get('AllowUserKeyImport'),
            csp_template=CodeSign.CSPTemplate(response_object.get('CSPTemplate')),
            disabled=response_object.get('Disabled'),
            dn=response_object.get('Dn'),
            encryption_key_algorithm=CodeSign.EnvironmentDetails(response_object.get('EncryptionKeyAlgorithm')),
            encryption_key_dn=response_object.get('EncryptionKeyDN'),
            expiration=response_object.get('Expiration'),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            ip_address_restriction=CodeSign.Items(response_object.get('IPAddressRestriction')),
            key_storage_location=CodeSign.EnvironmentDetails(response_object.get('KeyStorageLocation')),
            key_time_constraints=CodeSign.Items(response_object.get('KeyTimeConstraints')),
            key_use_flow_dn=response_object.get('KeyUseFlowDN'),
            max_uses=response_object.get('MaxUses'),
            per_user=response_object.get('PerUser'),
            signing_key_algorithm=CodeSign.EnvironmentDetails(response_object.get('SigningKeyAlgorithm')),
            signing_key_dn=response_object.get('SigningKeyDN'),
            template_dn=response_object.get('TemplateDN'),
        )

    @staticmethod
    def DotNetTemplate(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.DotNetTemplate(
            allow_user_key_import=response_object.get('AllowUserKeyImport'),
            disabled=response_object.get('Disabled'),
            dn=response_object.get('Dn'),
            expiration=CodeSign.InfoValue(response_object.get('Expiration')),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            key_algorithm=CodeSign.InfoValue(response_object.get('KeyAlgorithm')),
            key_container_dn=CodeSign.InfoValue(response_object.get('KeyContainerDN')),
            key_storage_location=CodeSign.InfoValue(response_object.get('KeyStorageLocation')),
            key_use_flow_dn=response_object.get('KeyUseFlowDN'),
            max_uses=CodeSign.InfoValue(response_object.get('MaxUses')),
            object_naming_pattern=response_object.get('ObjectNamingPattern'),
            per_user=response_object.get('PerUser'),
            type=response_object.get('Type'),
            visible_to=CodeSign.Items(response_object.get('VisibleTo')),
        )

    @staticmethod
    def DotNetEnvironment(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.DotNetEnvironment(
            allow_user_key_import=response_object.get('AllowUserKeyImport'),
            dot_net_template=CodeSign.DotNetTemplate(response_object.get('DotNetTemplate')),
            disabled=response_object.get('Disabled'),
            dn=response_object.get('Dn'),
            expiration=response_object.get('Expiration'),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            ip_address_restriction=response_object.get('IPAddressRestriction'),
            key_algorithm=CodeSign.EnvironmentDetails(response_object.get('KeyAlgorithm')),
            key_dn=response_object.get('KeyDn'),
            key_storage_location=CodeSign.EnvironmentDetails(response_object.get('KeyStorageLocation')),
            key_time_constraints=CodeSign.Items(response_object.get('KeyTimeConstraints')),
            key_use_flow_dn=response_object.get('KeyUseFlowDN'),
            max_uses=response_object.get('MaxUses'),
            per_user=response_object.get('PerUser'),
            template_dn=response_object.get('TemplateDN'),
        )

    @staticmethod
    def GPGTemplate(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.GPGTemplate(
            allow_user_key_import=response_object.get('AllowUserKeyImport'),
            authentication_key_algorithm=CodeSign.InfoValue(response_object.get('AuthenticationKeyAlgorithm')),
            dn=response_object.get('Dn'),
            email=CodeSign.InfoValue(response_object.get('Email')),
            encryption_key_algorithm=CodeSign.InfoValue(response_object.get('EncryptionKeyAlgorithm')),
            expiration=CodeSign.InfoValue(response_object.get('Expiration')),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            key_storage_location=CodeSign.InfoValue(response_object.get('KeyStorageLocation')),
            key_use_flow_dn=response_object.get('KeyUseFlowDN'),
            max_uses=CodeSign.InfoValue(response_object.get('MaxUses')),
            object_naming_pattern=response_object.get('ObjectNamingPattern'),
            per_user=response_object.get('PerUser'),
            read_only=response_object.get('ReadOnly'),
            real_name=CodeSign.InfoValue(response_object.get('RealName')),
            signing_key_algorithm=CodeSign.InfoValue(response_object.get('SigningKeyAlgorithm')),
            type=response_object.get('Type'),
            visible_to=CodeSign.Items(response_object.get('VisibleTo')),
        )
    
    @staticmethod
    def GPGEnvironment(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.GPGEnvironment(
            allow_user_key_import=response_object.get('AllowUserKeyImport'),
            authentication_key_algorithm=CodeSign.EnvironmentDetails(response_object.get('AuthenticationKeyAlgorithm')),
            custom_fields_attributes=[
                CodeSign.CustomFieldAttributes(cfa) 
                for cfa in response_object.get('CustomFieldsAttributes')
            ],
            dirty=response_object.get('Dirty'),
            dn=response_object.get('Dn'),
            email=response_object.get('Email'),
            encryption_key_algorithm=CodeSign.EnvironmentDetails(response_object.get('EncryptionKeyAlgorithm')),
            expiration=response_object.get('Expiration'),
            guid=response_object.get('Guid'),
            gpg_template=CodeSign.GPGTemplate(response_object.get('GpgTemplate')),
            id=response_object.get('Id'),
            ip_address_restriction=CodeSign.Items(response_object.get('IPAddressRestriction')),
            key_storage_location=CodeSign.EnvironmentDetails(response_object.get('KeyStorageLocation')),
            key_time_constraints=CodeSign.Items(response_object.get('KeyTimeConstraints')),
            key_use_flow_dn=response_object.get('KeyUseFlowDN'),
            max_uses=response_object.get('MaxUses'),
            per_user=response_object.get('PerUser'),
            real_name=CodeSign.EnvironmentDetails(response_object.get('RealName')),
            read_only=response_object.get('ReadOnly'),
            signing_key_algorithm=CodeSign.EnvironmentDetails(response_object.get('SigningKeyAlgorithm')),
            status=response_object.get('Status'),
            template_dn=response_object.get('TemplateDN'),
            type=response_object.get('Type'),
        )

    @staticmethod
    def KeyPairTemplate(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.KeyPairTemplate(
            allow_user_key_import=response_object.get('AllowUserKeyImport'),
            description=response_object.get('Description'),
            dirty=response_object.get('Dirty'),
            dn=response_object.get('Dn'),
            expiration=CodeSign.InfoValue(response_object.get('Expiration')),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            key_algorithm=CodeSign.InfoValue(response_object.get('KeyAlgorithm')),
            key_container_dn=response_object.get('KeyContainerDN'),
            key_storage_location=CodeSign.InfoValue(response_object.get('KeyStorageLocation')),
            max_uses=response_object.get('MaxUses'),
            type=response_object.get('Type'),
            validity_period=response_object.get('ValidityPeriod'),
            visible_to=CodeSign.Items(response_object.get('VisibleTo')),
        )
    
    @staticmethod
    def KeyPairEnvironment(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.KeyPairEnvironment(
            allow_user_key_import=response_object.get('AllowUserKeyImport'),
            custom_fields_attributes=[
                CodeSign.CustomFieldAttributes(cfa)
                for cfa in response_object.get('CustomFieldsAttributes')
            ],
            dn=response_object.get('Dn'),
            expiration=response_object.get('Expiration'),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            ip_address_restriction=CodeSign.Items(response_object.get('IPAddressRestriction')),
            key_algorithm=CodeSign.InfoValue(response_object.get('KeyAlgorithm')),
            key_dn=response_object.get('KeyDn'),
            key_pair_template=CodeSign.KeyPairTemplate(response_object.get('KeyPairTemplate')),
            key_storage_location=CodeSign.EnvironmentDetails(response_object.get('KeyStorageLocation')),
            key_time_constraints=CodeSign.Items(response_object.get('KeyTimeConstraints')),
            key_use_flow_dn=response_object.get('KeyUseFlowDn'),
            status=response_object.get('Status'),
            template_dn=response_object.get('TemplateDN'),
            type=response_object.get('Type'),
        )

    @staticmethod
    def GlobalConfiguration(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.GlobalConfiguration(
            approved_key_storage_locations=CodeSign.KeyStorageLocations(
                response_object.get('ApprovedKeyStorageLocations')
            ),
            available_key_storage_locations=CodeSign.KeyStorageLocations(
                response_object.get('AvailableKeyStorageLocations')
            ),
            default_ca_container=response_object.get('DefaultCAContainer'),
            default_certificate_container=response_object.get('DefaultCertificateContainer'),
            default_credential_container=response_object.get('DefaultCredentialContainer'),
            key_use_timeout=response_object.get('KeyUseTimeout'),
            project_description_tooltip=response_object.get('ProjectDescriptionTooltip'),
            request_in_progress_message=response_object.get('RequestInProgressMessage'),
        )

    @staticmethod
    def Project(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.Project(
            application_dns=CodeSign.Items(response_object.get('ApplicationDNs')),
            applications=[CodeSign.Application(a) for a in response_object.get('Applications')],
            auditors=CodeSign.Items(response_object.get('Auditors')),
            certificate_environments=[
                CodeSign.CertificateEnvironment(ce)
                for ce in response_object.get('CertificateEnvironments')
            ],
            collections=[
                CodeSign.SignApplicationCollection(c)
                for c in response_object.get('Collections')
            ],
            created_on=from_date_string(response_object.get('CreatedOn')),
            description=response_object.get('Description'),
            dn=response_object.get('Dn'),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            key_use_approvers=CodeSign.Items(response_object.get('KeyUseApprovers')),
            key_users=CodeSign.Items(response_object.get('KeyUsers')),
            owners=CodeSign.Items(response_object.get('Owners')),
            status=response_object.get('Status'),
        )

    @staticmethod
    def Rights(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        rights = response_object.get('Rights')
        return codesign.Rights(
            none=rights == 0,
            admin=rights & 1 != 0,
            use=rights & 2 != 0,
            audit=rights & 4 != 0,
            owner=rights & 8 != 0,
            project_approval=rights & 16 != 0,
            application_admin=rights & 32 != 0,
            approve_use=rights & 64 != 0,
        )

    @staticmethod
    def SignApplicationCollection(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return codesign.SignApplicationCollection(
            description=response_object.get('Description'),
            dn=response_object.get('Dn'),
            guid=response_object.get('Guid'),
            hash=response_object.get('Hash'),
            id=response_object.get('Id'),
            location=response_object.get('Location'),
            permitted_argument_pattern=response_object.get('PermittedArgumentPattern'),
            signatory_issuer=response_object.get('SignatoryIssuer'),
            signatory_subject=response_object.get('SignatorySubject'),
            size=response_object.get('Size'),
            version=response_object.get('Version'),
        )



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.flow_action_config_read_approvers import FlowActionConfigReadApproversAttributes


class FlowActionTeamOwnerApprovalAttributes(FlowActionConfigReadApproversAttributes, metaclass=IterableMeta):
	__config_class__ = "Flow Action Team Owner Approval"



from pytpp.attributes._helper import IterableMeta, Attribute


class ClientPortalBaseAttributes(metaclass=IterableMeta):
	__config_class__ = "Client Portal Base"
	client_field_mapping = Attribute('Client Field Mapping', min_version='15.2')
	client_portal_access_enabled = Attribute('Client Portal Access Enabled', min_version='15.2')



from pytpp.properties.resultcodes import ResultCodes
from pytpp.properties.response_objects.dataclasses import metadata


class Metadata:
    @staticmethod
    def Result(code: int):
        return metadata.Result(
            code=code,
            metadata_result=ResultCodes.Metadata.get(code, 'Unknown'),
        )

    @staticmethod
    def Item(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return metadata.Item(
            allowed_characters=response_object.get('AllowedCharacters'),
            allowed_values=response_object.get('AllowedValues'),
            category=response_object.get('Category'),
            classes=response_object.get('Classes'),
            config_attribute=response_object.get('ConfigAttribute'),
            date_only=response_object.get('DateOnly'),
            default_values=response_object.get('DefaultValues'),
            display_after=response_object.get('DisplayAfter'),
            dn=response_object.get('DN'),
            error_message=response_object.get('ErrorMessage'),
            guid=response_object.get('Guid'),
            help=response_object.get('Help'),
            label=response_object.get('Label'),
            localization_table=response_object.get('LocalizationTable'),
            localized_help=response_object.get('LocalizedHelp'),
            localized_label=response_object.get('LocalizedLabel'),
            localized_set=response_object.get('LocalizedSet'),
            mandatory=response_object.get('Mandatory'),
            name=response_object.get('Name'),
            mask=response_object.get('Mask'),
            maximum_length=response_object.get('MaximumLength'),
            minimum_length=response_object.get('MinimumLength'),
            policyable=response_object.get('Policyable'),
            regular_expression=response_object.get('RegularExpression'),
            render_hidden=response_object.get('RenderHidden'),
            render_read_only=response_object.get('RenderReadOnly'),
            single=response_object.get('Single'),
            time_only=response_object.get('TimeOnly'),
            type=response_object.get('Type'),
        )

    @staticmethod
    def Data(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return metadata.Data(
            key=Metadata.Item(response_object.get('Key')),
            value=response_object.get('Value'),
        )

    @staticmethod
    def PolicyItem(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return metadata.PolicyItem(
            key=response_object.get('Key'),
            value=[Metadata.Item(value) for value in response_object.get('Value')],
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.discovery_statistics import DiscoveryStatisticsAttributes
from pytpp.attributes.schedule_base import ScheduleBaseAttributes
from pytpp.attributes.top import TopAttributes


class DiscoveryAttributes(DiscoveryStatisticsAttributes, ScheduleBaseAttributes, TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Discovery"
	address_parsing_errors = Attribute('Address Parsing Errors', min_version='15.4')
	address_range = Attribute('Address Range')
	automatically_import = Attribute('Automatically Import')
	certificate_instances_found = Attribute('Certificate Instances Found')
	certificate_location_dn = Attribute('Certificate Location DN')
	certificates_already_known = Attribute('Certificates Already Known')
	certificates_excluded = Attribute('Certificates Excluded')
	completed = Attribute('Completed')
	completed_assignments = Attribute('Completed Assignments')
	configuration = Attribute('Configuration')
	connect_excluded = Attribute('Connect Excluded')
	creation_date = Attribute('Creation Date')
	device_location_dn = Attribute('Device Location DN')
	discovery_exclusion_dn = Attribute('Discovery Exclusion DN')
	in_progress = Attribute('In Progress')
	keys_already_known = Attribute('Keys Already Known')
	keys_excluded = Attribute('Keys Excluded')
	last_update = Attribute('Last Update', min_version='16.2')
	new_certificate_instances_found = Attribute('New Certificate Instances Found')
	new_certificates_found = Attribute('New Certificates Found')
	new_ssh_servers_found = Attribute('New SSH Servers Found', min_version='16.3')
	placement_preview = Attribute('Placement Preview')
	placement_rule = Attribute('Placement Rule')
	placement_summary = Attribute('Placement Summary')
	priority = Attribute('Priority')
	protection_key = Attribute('Protection Key')
	report_dn = Attribute('Report DN')
	resolve_host = Attribute('Resolve Host')
	ssh_servers_found = Attribute('SSH Servers Found', min_version='16.3')
	started = Attribute('Started')
	status = Attribute('Status')
	total_assignments = Attribute('Total Assignments')
	window_days_of_month = Attribute('Window Days of Month')
	window_days_of_week = Attribute('Window Days of Week')
	window_end = Attribute('Window End')
	window_start = Attribute('Window Start')
	work_units = Attribute('Work Units')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.agent_discovery_base import AgentDiscoveryBaseAttributes


class AgentCertificateDiscoveryAttributes(AgentDiscoveryBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Agent Certificate Discovery"



import warnings
from datetime import datetime
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.tools.helpers.date_converter import from_date_string


class _Authorize(API):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Authorize')

        self.Certificate = self._Certificate(api_obj=api_obj)
        self.Integrated = self._Integrated(api_obj=api_obj)
        self.OAuth = self._OAuth(api_obj=api_obj)
        self.Token = self._Token(api_obj=api_obj)
        self.Verify = self._Verify(api_obj=api_obj)

    def post(self, username, password):
        """
        This POST method is written differently in order to effectively omit the password from being logged.
        """
        self._log_api_deprecated_warning(alternate_api=self.OAuth._url)
        warnings.warn('Authorizing to TPP with only a username and password is being deprecated. '
                      'Refer to product documentation on using OAuth authentication.')
        body = {
            "Username": username,
            "Password": password
        }

        class _Response(APIResponse):
            def __init__(self, response):
                super().__init__(response=response)

            @property
            @api_response_property()
            def token(self) -> str:
                return self._from_json('APIKey')

        return _Response(
            response=self._post(data=body))

    class _Certificate(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='Authorize/Certificate')
            self._url = self._url.replace('vedsdk', 'vedauth')

        def post(self, client_id: str, scope: str):
            body = {
                'client_id': client_id,
                'scope': scope
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def access_token(self) -> str:
                    return self._from_json('access_token')

                @property
                @api_response_property()
                def expires(self):
                    return datetime.fromtimestamp(self._from_json('expires'))

                @property
                @api_response_property()
                def expires_in(self) -> int:
                    return self._from_json(key='expires_in')

                @property
                @api_response_property()
                def identity(self) -> str:
                    return self._from_json('identity')

                @property
                @api_response_property()
                def refresh_token(self) -> str:
                    return self._from_json('refresh_token')

                @property
                @api_response_property()
                def refresh_until(self):
                    return datetime.fromtimestamp(self._from_json(key='refresh_until'))

                @property
                @api_response_property()
                def scope(self) -> str:
                    return self._from_json('scope')

                @property
                @api_response_property()
                def token_type(self) -> str:
                    return self._from_json('token_type')

            return _Response(response=self._post(data=body))

    class _Integrated(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='Authorize/Integrated')
            self._url = self._url.replace('vedsdk', 'vedauth')

        def post(self, client_id: str, password: str, scope: str, username: str, state: str = None):
            body = {
                'client_id': client_id,
                'password' : password,
                'scope'    : scope,
                'username' : username,
                'state'    : state
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def access_token(self) -> str:
                    return self._from_json('access_token')

                @property
                @api_response_property()
                def expires(self):
                    return datetime.fromtimestamp(self._from_json('expires'))

                @property
                @api_response_property()
                def expires_in(self) -> int:
                    return self._from_json(key='expires_in')

                @property
                @api_response_property()
                def identity(self) -> str:
                    return self._from_json('identity')

                @property
                @api_response_property()
                def refresh_token(self) -> str:
                    return self._from_json('refresh_token')

                @property
                @api_response_property()
                def refresh_until(self):
                    return datetime.fromtimestamp(self._from_json(key='refresh_until'))

                @property
                @api_response_property()
                def scope(self) -> str:
                    return self._from_json('scope')

                @property
                @api_response_property()
                def token_type(self) -> str:
                    return self._from_json('token_type')

            return _Response(response=self._post(data=body))

    class _OAuth(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Authorize/OAuth')
            self._url = self._url.replace('vedsdk', 'vedauth')

        def post(self, client_id: str, password: str, scope: str, username: str, state: str = None):
            body = {
                'client_id': client_id,
                'password': password,
                'scope': scope,
                'username': username,
                'state': state
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def access_token(self) -> str:
                    return self._from_json('access_token')

                @property
                @api_response_property()
                def expires(self) -> str:
                    return self._from_json('expires')

                @property
                @api_response_property()
                def identity(self) -> str:
                    return self._from_json('identity')

                @property
                @api_response_property()
                def refresh_token(self) -> str:
                    return self._from_json('refresh_token')

                @property
                @api_response_property()
                def scope(self) -> str:
                    return self._from_json('scope')

                @property
                @api_response_property()
                def token_type(self) -> str:
                    return self._from_json('token_type')

            return _Response(
                response=self._post(data=body))

    class _Token(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Authorize/Token')
            self._url = self._url.replace('vedsdk', 'vedauth')

        def post(self, client_id: str, refresh_token: str):
            body = {
                'client_id': client_id,
                'refresh_token': refresh_token
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def access_token(self) -> str:
                    return self._from_json('access_token')

                @property
                @api_response_property()
                def expires(self) -> str:
                    return self._from_json('expires')

                @property
                @api_response_property()
                def refresh_token(self) -> str:
                    return self._from_json('refresh_token')

                @property
                @api_response_property()
                def refresh_until(self) -> str:
                    return self._from_json('refresh_until')

                @property
                @api_response_property()
                def scope(self) -> str:
                    return self._from_json('scope')

                @property
                @api_response_property()
                def token_type(self) -> str:
                    return self._from_json('token_type')

            return _Response(response=self._post(data=body))

    class _Verify(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='Authorize/Verify')
            self._url = self._url.replace('vedsdk', 'vedauth')

        def get(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def access_issued_on(self):
                    return from_date_string(self._from_json(key='access_issued_on'))

                @property
                @api_response_property()
                def access_issued_on_ISO8601(self):
                    return from_date_string(self._from_json(key='access_issued_on_ISO8601'))

                @property
                @api_response_property()
                def access_issued_on_unix_time(self):
                    return datetime.fromtimestamp(self._from_json(key='access_issued_on_unix_time'))

                @property
                @api_response_property()
                def application(self) -> str:
                    return self._from_json(key='application')

                @property
                @api_response_property()
                def expires(self):
                    return from_date_string(self._from_json(key='expires'))

                @property
                @api_response_property()
                def expires_ISO8601(self):
                    return from_date_string(self._from_json(key='expires_ISO8601'))

                @property
                @api_response_property()
                def expires_unix_time(self):
                    return datetime.fromtimestamp(self._from_json(key='expires_unix_time'))

                @property
                @api_response_property()
                def grant_issued_on(self):
                    return from_date_string(self._from_json(key='grant_issued_on'))

                @property
                @api_response_property()
                def grant_issued_on_ISO8601(self):
                    return from_date_string(self._from_json(key='grant_issued_on_ISO8601'))

                @property
                @api_response_property()
                def grant_issued_on_unix_time(self):
                    return datetime.fromtimestamp(self._from_json(key='grant_issued_on_unix_time'))

                @property
                @api_response_property()
                def identity(self) -> str:
                    return self._from_json(key='identity')

                @property
                @api_response_property()
                def scope(self) -> str:
                    return self._from_json(key='scope')

                @property
                @api_response_property()
                def valid_for(self) -> int:
                    return self._from_json(key='valid_for')

            return _Response(response=self._get())



import time
from typing import List, Dict
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.credential import Credential
from pytpp.tools.helpers.date_converter import from_date_string


class _Credentials:
    def __init__(self, api_obj):
        self.Adaptable = self._Adaptable(api_obj=api_obj)
        self.Connector = self._Connector(api_obj=api_obj)
        self.Create = self._Create(api_obj=api_obj)
        self.Delete = self._Delete(api_obj=api_obj)
        self.Enumerate = self._Enumerate(api_obj=api_obj)
        self.Rename = self._Rename(api_obj=api_obj)
        self.Retrieve = self._Retrieve(api_obj=api_obj)
        self.Update = self._Update(api_obj=api_obj)
        self.CyberArk = self._CyberArk(api_obj=api_obj)

    class _Adaptable:
        def __init__(self, api_obj):
            self.Update = self._Update(api_obj=api_obj)
        
        class _Update(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/Credentials/Adaptable/Update')
            
            def post(self, credential_path: str, credential_type: str, connector_name: str,
                     custom_fields: List[Dict[str, str]]):
                body = {
                    'CredentialPath': credential_path,
                    'CredentialType': credential_type,
                    'ConnectorName': connector_name,
                    'CustomFields': custom_fields
                }
                
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def result(self):
                        return Credential.Result(self._from_json(key='Result'))
                        
                return _Response(response=self._post(data=body))

    class _Connector:
        def __init__(self, api_obj):
            self.Adaptable = self._Adaptable(api_obj=api_obj)

        class _Adaptable(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/Credentials/Connector/Adaptable')

            def post(self, connector_name: str, powershell_script: str, service_address: str,
                     service_credential: str, allowed_identities: List[str] = None,
                     description: str = None):
                body = {
                    'AllowedIdentities': allowed_identities,
                    'ConnectorName': connector_name,
                    'Description': description,
                    'PowershellScript': powershell_script,
                    'ServiceAddress': service_address,
                    'ServiceCredential': service_credential
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def succcess(self) -> bool:
                        return self._from_json(key='Succcess')

                return _Response(response=self._post(data=body))

            def Guid(self, guid: str):
                return self._Guid(api_obj=self._api_obj, guid=guid)

            class _Guid(API):
                def __init__(self, api_obj, guid: str):
                    super().__init__(api_obj=api_obj, url=f'/Credentials/Connector/Adaptable/{guid}')

                def delete(self):
                    class _Response(APIResponse):
                        def __init__(self, response):
                            super().__init__(response=response)

                        @property
                        @api_response_property()
                        def success(self) -> bool:
                            return self._from_json(key='Success')

                    return _Response(response=self._delete())

                def get(self):
                    class _Response(APIResponse):
                        def __init__(self, response):
                            super().__init__(response=response)

                        @property
                        @api_response_property()
                        def allowed_identities(self) -> List[str]:
                            return self._from_json(key='AllowedIdentities')

                        @property
                        @api_response_property()
                        def powershell_script(self) -> str:
                            return self._from_json(key='PowershellScript')

                        @property
                        @api_response_property()
                        def service_address(self) -> str:
                            return self._from_json(key='ServiceAddress')

                        @property
                        @api_response_property()
                        def service_credential(self) -> str:
                            return self._from_json(key='ServiceCredential')

                        @property
                        @api_response_property()
                        def success(self) -> bool:
                            return self._from_json(key='Success')

                    return _Response(response=self._get())

                def put(self, connector_name: str = None, powershell_script: str = None,
                        service_address: str = None, service_credential: str = None,
                        allowed_identities: List[str] = None, description: str = None):
                    body = {
                        'AllowedIdentities': allowed_identities,
                        'ConnectorName'    : connector_name,
                        'Description'      : description,
                        'PowershellScript' : powershell_script,
                        'ServiceAddress'   : service_address,
                        'ServiceCredential': service_credential
                    }

                    class _Response(APIResponse):
                        def __init__(self, response):
                            super().__init__(response=response)

                        @property
                        @api_response_property()
                        def succcess(self) -> bool:
                            return self._from_json(key='Succcess')

                    return _Response(response=self._put(data=body))

    class _Create(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Credentials/Create')

        def post(self, credential_path: str, friendly_name: str, values: list, password: str = None, description: str = None,
                 encryption_key: str = None, shared: bool = False, expiration: int = None, contact: list = None):
            body = {
                'CredentialPath': credential_path,
                'Password': password,
                'FriendlyName': friendly_name,
                'Values': values,
                'Expiration': f'/Date({expiration})/',
                'Description': description,
                'EncryptionKey': encryption_key,
                'Shared': shared,
                'Contact': contact
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Credential.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _Delete(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Credentials/Delete')

        def post(self, credential_path: str):
            body = {
                'CredentialPath': credential_path
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Credential.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _Enumerate(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Credentials/Enumerate')

        def post(self, credential_path: str, pattern: str = None, recursive: bool = False):
            body = {
                'CredentialPath': credential_path,
                'Recursive': recursive,
                'Pattern': pattern
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Credential.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def credential_infos(self):
                    return [Credential.CredentialInfo(cred_info) for cred_info in self._from_json(key='CredentialInfos')]

            return _Response(response=self._post(data=body))

    class _Rename(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Credentials/Rename')

        def post(self, credential_path: str, new_credential_path: str):
            body = {
                'CredentialPath': credential_path,
                'NewCredentialPath': new_credential_path
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Credential.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _Retrieve(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Credentials/Retrieve')

        def post(self, credential_path: str):
            body = {
                'CredentialPath': credential_path
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def classname(self) -> str:
                    return self._from_json(key='Classname')

                @property
                @api_response_property()
                def description(self) -> str:
                    return self._from_json(key='Description')

                @property
                @api_response_property()
                def expiration(self):
                    return from_date_string(self._from_json(key='Expiration'))

                @property
                @api_response_property()
                def friendly_name(self) -> str:
                    return self._from_json(key='FriendlyName')

                @property
                @api_response_property()
                def result(self):
                    return Credential.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def values(self):
                    return [Credential.NameTypeValue(ntv) for ntv in self._from_json(key='Values')]

            return _Response(response=self._post(data=body))

    class _Update(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Credentials/Update')

        def post(self, credential_path: str, friendly_name: str, values: list, description: str = None,
                 encryption_key: str = None, shared: bool = False, expiration: int = None, contact: list = None):
            body = {
                'CredentialPath': credential_path,
                'FriendlyName': friendly_name,
                'Values': values,
                'Description': description,
                'EncryptionKey': encryption_key,
                'Shared': shared,
                'Contact': contact
            }

            if expiration:
                exp_date = expiration
            else:
                # Expire in 10 years.
                exp_date = int((time.time() + (60 * 60 * 24 * 365 * 10)) * 1000)

            body.update({'Expiration': r'/Date(%s)/' % exp_date})
            
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return Credential.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _CyberArk:
        def __init__(self, api_obj):
            self.Create = self._Create(api_obj=api_obj)
            self.Update = self._Update(api_obj=api_obj)

        class _Create(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/Credentials/CyberArk/Create')

            def post(self, cyber_ark_username: str, cyber_ark_password: str, username: str, app_id: str, safe_name: str,
                     folder_name: str, account_name: str, credentials_path: str):
                body = {
                    'CyberArkUsername': cyber_ark_username,
                    'CyberArkPassword': cyber_ark_password,
                    'Username': username,
                    'AppID': app_id,
                    'SafeName': safe_name,
                    'FolderName': folder_name,
                    'AccountName': account_name,
                    'CredentialsPath': credentials_path
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def result(self):
                        return Credential.Result(self._from_json(key='Result'))

                return _Response(response=self._post(data=body))

        class _Update(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/Credentials/CyberArk/Update')

            def post(self, cyber_ark_username: str, cyber_ark_password: str, username: str, app_id: str, safe_name: str,
                     folder_name: str, account_name: str, credentials_path: str):
                body = {
                    'CyberArkUsername': cyber_ark_username,
                    'CyberArkPassword': cyber_ark_password,
                    'Username': username,
                    'AppID': app_id,
                    'SafeName': safe_name,
                    'FolderName': folder_name,
                    'AccountName': account_name,
                    'CredentialsPath': credentials_path
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def result(self):
                        return Credential.Result(self._from_json(key='Result'))

                return _Response(response=self._post(data=body))



from pytpp.properties.resultcodes import ResultCodes
from pytpp.properties.response_objects.dataclasses import config_schema


class ConfigSchema:
    @staticmethod
    def Result(code: int):
        return config_schema.Result(
            code=code,
            config_result=ResultCodes.Config.get(code, 'Unknown'),
        )

    @staticmethod
    def AttributeDefinition(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return config_schema.AttributeDefinition(
            name=response_object.get('Name'),
            syntax=response_object.get('Syntax'),
        )

    @staticmethod
    def ClassDefinition(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return config_schema.ClassDefinition(
            containment_names=response_object.get('ContainmentNames'),
            containment_sub_names=response_object.get('ContainmentSubNames'),
            mandatory_names=response_object.get('MandatoryNames'),
            name=response_object.get('Name'),
            naming_names=response_object.get('NamingNames'),
            optional_names=response_object.get('OptionalNames'),
            super_class_names=response_object.get('SuperClassNames'),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class CiscoCSMAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "CiscoCSM"
	create_chain_trustpoints = Attribute('Create Chain Trustpoints')
	create_new_trustpoint = Attribute('Create New Trustpoint')
	enable_credential = Attribute('Enable Credential')
	generate_new_key = Attribute('Generate New Key')
	key_created = Attribute('Key Created')
	key_pair_name = Attribute('Key Pair Name')
	trustpoint_name = Attribute('Trustpoint Name')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.organization import OrganizationAttributes


class LayoutRuleContainerAttributes(OrganizationAttributes, metaclass=IterableMeta):
	__config_class__ = "Layout Rule Container"
